import copy
import logging
from unittest.mock import MagicMock, mock_open, patch

import pytest

from perun.connector import perun_openapi
from perun.connector.adapters.PerunRpcAdapter import PerunRpcAdapter
from perun.connector.models.Facility import Facility
from perun.connector.models.Group import Group
from perun.connector.models.Member import Member
from perun.connector.models.MemberStatusEnum import MemberStatusEnum
from perun.connector.models.Resource import Resource
from perun.connector.models.User import User
from perun.connector.models.UserExtSource import UserExtSource
from perun.connector.models.VO import VO
from perun.connector.perun_openapi.exceptions import ApiException
from perun.connector.utils.ConfigStore import ConfigStore


class HttpResponse:
    def __init__(self, input_data: str):
        self.status = None
        self.data = input_data
        self.reason = None

    def getheaders(self):
        return "headers"


CONFIG = ConfigStore.get_openapi_config()

COOKIE_FILEPATH = CONFIG.get("cookie_filepath", "cookie.txt")

# sample groups - external representation (Devel)
TEST_GROUP_EXTERNAL_REPRESENTATION_1 = {
    "id": 1,
    "vo_id": 10,
    "uuid": "sample-uuid-value",
    "name": "sample:group:name",
    "description": "This is a sample group",
}

TEST_GROUP_EXTERNAL_REPRESENTATION_2 = {
    "id": 2,
    "vo_id": 10,
    "uuid": "sample-uuid-value-2",
    "name": "sample:group:name:specific",
    "description": "This is a sample sub-group",
}

# sample groups - internal representation

TEST_VO = VO(62, "CESNET e-infrastruktura", "einfra")
TEST_MEMBER_INTERNAL_REPRESENTATION = Member(5, TEST_VO, "VALID")
TEST_MEMBER_EXTERNAL_REPRESENTATION = {"id": 5, "vo_id": TEST_VO.id, "status": "VALID"}
TEST_USER = User(10, "John Doe")
SAMPLE_SHORT_GROUP_NAME = "short_name"

TEST_GROUP_INTERNAL_REPRESENTATION_1 = Group(
    TEST_GROUP_EXTERNAL_REPRESENTATION_1["id"],
    TEST_VO,
    TEST_GROUP_EXTERNAL_REPRESENTATION_1["uuid"],
    TEST_GROUP_EXTERNAL_REPRESENTATION_1["name"],
    f"{SAMPLE_SHORT_GROUP_NAME}:" f'{TEST_GROUP_EXTERNAL_REPRESENTATION_1["name"]}',
    TEST_GROUP_EXTERNAL_REPRESENTATION_1["description"],
)

TEST_GROUP_INTERNAL_REPRESENTATION_2 = Group(
    TEST_GROUP_EXTERNAL_REPRESENTATION_2["id"],
    TEST_VO,
    TEST_GROUP_EXTERNAL_REPRESENTATION_2["uuid"],
    TEST_GROUP_EXTERNAL_REPRESENTATION_2["name"],
    f"{SAMPLE_SHORT_GROUP_NAME}:" f'{TEST_GROUP_EXTERNAL_REPRESENTATION_2["name"]}',
    TEST_GROUP_EXTERNAL_REPRESENTATION_2["description"],
)

# sample Facility
TEST_RP_IDENTIFIER_1 = "test rp identifier"
TEST_RP_IDENTIFIER_2 = "test rp identifier 2"

TEST_EXTERNAL_FACILITY_1 = {
    "id": 1,
    "name": "test name",
    "description": "test description",
}
TEST_EXTERNAL_FACILITY_2 = {
    "id": 2,
    "name": "test name 2",
    "description": "test description 2",
}

TEST_EXTERNAL_FACILITY_WITH_ATTRS_1 = [
    {
        "facility": {
            "id": 1,
            "name": "test name",
            "description": "test description",
        },
        "attributes": [
            {
                "baseFriendlyName": "rpIdentifier",
                "base_friendly_name": "rpIdentifier",
                "bean_name": "Attribute",
                "description": "Unique identifier of RP",
                "displayName": "RP Identifier",
                "display_name": "RP Identifier",
                "entity": "facility",
                "friendlyName": "rpIdentifier",
                "friendlyNameParameter": "",
                "friendly_name": "rpIdentifier",
                "friendly_name_parameter": "",
                "namespace": "urn:perun:facility:attribute-def:def",
                "type": "java.lang.String",
                "unique": True,
                "value": TEST_RP_IDENTIFIER_1,
                "writable": True,
            }
        ],
    }
]

TEST_SINGLE_PERUN_FACILITY = [TEST_EXTERNAL_FACILITY_1]
TEST_MULTIPLE_PERUN_FACILITIES = [TEST_EXTERNAL_FACILITY_1, TEST_EXTERNAL_FACILITY_2]

TEST_INTERNAL_FACILITY_1 = Facility(
    TEST_EXTERNAL_FACILITY_1["id"],
    TEST_EXTERNAL_FACILITY_1["name"],
    TEST_EXTERNAL_FACILITY_1["description"],
    TEST_RP_IDENTIFIER_1,
    {},
)

TEST_INTERNAL_FACILITY_2 = Facility(
    TEST_EXTERNAL_FACILITY_2["id"],
    TEST_EXTERNAL_FACILITY_2["name"],
    TEST_EXTERNAL_FACILITY_2["description"],
    TEST_RP_IDENTIFIER_2,
    {},
)

MULTIPLE_ATTRS_ERROR_TEXT = (
    f"There is more than one facility with rpID '" f"{TEST_RP_IDENTIFIER_1}'."
)
NO_ATTRS_ERROR_TEXT = f"No facility with rpID '{TEST_RP_IDENTIFIER_1}' found."

COOKIE = "PERUNSESSION=TEST_COOKIE;"
RESPONSE_HEADER_COOKIE = {"Set-Cookie": COOKIE}


def get_entity_specific_attributes(attribute_type: str):
    generated_attrs = []
    attr_names = []
    attr_values = []
    attr_namespaces = []

    if attribute_type == "user":
        attr_names = [
            "perunUserAttribute_phone",
            "perunUserAttribute_groupNames",
            "perunUserAttribute_preferredLanguage",
        ]
        attr_values = [
            None,
            ["group 1 name", "group 2 name"],
            "sample attribute value 3",
        ]
        attr_namespaces = [
            "urn:perun:user:attribute-def:def",
            "urn:perun:user:attribute-def:virt",
            "urn:perun:user:attribute-def:def",
        ]

    elif attribute_type == "vo":
        attr_names = ["perunVoAttribute_aup"]
        attr_values = ["sample value"]
        attr_namespaces = ["urn:perun:vo:attribute-def:def"]

    elif attribute_type == "entityless":
        attr_names = [
            "perunEntitylessAttribute_orgAups",
            "perunEntitylessAttribute_orgAups",
        ]
        attr_values = [None, "sample attribute value"]
        attr_namespaces = [
            "urn:perun:entityless:attribute-def:def",
            "urn:perun:entityless:attribute-def:def",
        ]

    elif attribute_type == "facility":
        attr_names = ["perunFacilityAttr_checkGroupMembership"]
        attr_values = ["test value"]
        attr_namespaces = ["urn:perun:facility:attribute-def:def"]

    for index, attr_name in enumerate(attr_names):
        generated_attrs.append(
            {
                "id": index,
                "name": attr_name,
                "namespace": attr_namespaces[index],
                "display_name": f"display name for {attr_name}",
                "type": "some type",
                "value": attr_values[index],
                "friendlyName": attr_name.split("_")[-1],
            }
        )

    return generated_attrs


@patch(
    "perun.connector.perun_openapi.api.users_manager_api.UsersManagerApi"
    ".get_user_by_ext_source_name_and_ext_login"
)
@patch("builtins.open", new_callable=mock_open, read_data="")
def test_get_perun_user_with_middle_name_and_title_before_after(
    mock_cookie, mock_request_1
):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    user_with_middle_name_title_before_after = {
        "id": 10,
        "title_before": "Ing.",
        "first_name": "John",
        "middle_name": "Frederick",
        "last_name": "Doe",
        "title_after": "Ph.D.",
    }
    perun_openapi.api.users_manager_api.UsersManagerApi.get_user_by_ext_source_name_and_ext_login = MagicMock(
        return_value=(
            user_with_middle_name_title_before_after,
            200,
            RESPONSE_HEADER_COOKIE,
        )
    )

    expected_user = User(10, "Ing. John Frederick Doe Ph.D.")
    result_user = adapter.get_perun_user("10", ["John Doe"])

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_user == expected_user


@patch(
    "perun.connector.perun_openapi.api.users_manager_api.UsersManagerApi"
    ".get_user_by_ext_source_name_and_ext_login"
)
@patch("builtins.open", new_callable=mock_open, read_data="")
def test_get_perun_user_without_middle_name_and_title_before_after(
    mock_cookie, mock_request_1
):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    user_without_middle_name_title_before_after = {
        "id": 10,
        "title_before": "Ing.",
        "first_name": "John",
        "middle_name": None,
        "last_name": "Doe",
        "title_after": "Ph.D.",
    }
    perun_openapi.api.users_manager_api.UsersManagerApi.get_user_by_ext_source_name_and_ext_login = MagicMock(
        return_value=(
            user_without_middle_name_title_before_after,
            200,
            RESPONSE_HEADER_COOKIE,
        )
    )

    expected_user = User(10, "Ing. John Doe Ph.D.")
    result_user = adapter.get_perun_user("10", ["John Doe"])

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_user == expected_user


@patch(
    "perun.connector.perun_openapi.api.users_manager_api.UsersManagerApi"
    ".get_user_by_ext_source_name_and_ext_login"
)
@patch("builtins.open", new_callable=mock_open)
def test_user_without_middle_name_and_title_before(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    user_without_middle_name_title_before = {
        "id": 10,
        "title_before": "Ing.",
        "first_name": "John",
        "middle_name": None,
        "last_name": "Doe",
        "title_after": None,
    }

    perun_openapi.api.users_manager_api.UsersManagerApi.get_user_by_ext_source_name_and_ext_login = MagicMock(
        return_value=(
            user_without_middle_name_title_before,
            200,
            RESPONSE_HEADER_COOKIE,
        )
    )

    expected_user = User(10, "Ing. John Doe")
    result_user = adapter.get_perun_user("10", ["John Doe"])

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_user == expected_user


@patch(
    "perun.connector.perun_openapi.api.users_manager_api.UsersManagerApi"
    ".get_user_by_ext_source_name_and_ext_login"
)
@patch("builtins.open", new_callable=mock_open)
def test_user_without_middle_name_and_title_after(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    user_without_middle_name_title_after = {
        "id": 10,
        "title_before": None,
        "first_name": "John",
        "middle_name": None,
        "last_name": "Doe",
        "title_after": "Ph.D.",
    }

    perun_openapi.api.users_manager_api.UsersManagerApi.get_user_by_ext_source_name_and_ext_login = MagicMock(
        return_value=(user_without_middle_name_title_after, 200, RESPONSE_HEADER_COOKIE)
    )

    expected_user = User(10, "John Doe Ph.D.")
    result_user = adapter.get_perun_user("10", ["John Doe"])

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_user == expected_user


@patch(
    "perun.connector.perun_openapi.api.users_manager_api.UsersManagerApi"
    ".get_user_by_ext_source_name_and_ext_login"
)
@patch("builtins.open", new_callable=mock_open)
def test_user_without_middle_name_and_titles(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    user_without_middle_name_and_titles = {
        "id": 10,
        "title_before": None,
        "first_name": "John",
        "middle_name": None,
        "last_name": "Doe",
        "title_after": None,
    }

    perun_openapi.api.users_manager_api.UsersManagerApi.get_user_by_ext_source_name_and_ext_login = MagicMock(
        return_value=(user_without_middle_name_and_titles, 200, RESPONSE_HEADER_COOKIE)
    )

    expected_user = User(10, "John Doe")
    result_user = adapter.get_perun_user("10", ["John Doe"])

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_user == expected_user


# get_get_member_groups tests
@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_attribute"
)
@patch(
    "perun.connector.perun_openapi.api.groups_manager_api.GroupsManagerApi"
    ".get_all_member_groups"
)
@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
@patch("perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id")
@patch("builtins.open", new_callable=mock_open)
def test_get_member_groups_found_member_groups(
    mock_cookie, mock_request_1, mock_request_2, mock_request_3, mock_request_4
):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_attribute = (
        MagicMock(
            return_value=(
                {"value": SAMPLE_SHORT_GROUP_NAME},
                200,
                RESPONSE_HEADER_COOKIE,
            )
        )
    )

    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id = MagicMock(
        return_value=(TEST_VO, 200, RESPONSE_HEADER_COOKIE)
    )

    perun_openapi.api.groups_manager_api.GroupsManagerApi.get_all_member_groups = (
        MagicMock(
            return_value=(
                [
                    TEST_GROUP_EXTERNAL_REPRESENTATION_1,
                    TEST_GROUP_EXTERNAL_REPRESENTATION_2,
                ],
                200,
                RESPONSE_HEADER_COOKIE,
            )
        )
    )

    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = (
        MagicMock(
            return_value=(
                TEST_MEMBER_INTERNAL_REPRESENTATION.__dict__,
                200,
                RESPONSE_HEADER_COOKIE,
            )
        )
    )

    result_groups = adapter.get_member_groups(user=TEST_USER, vo=TEST_VO)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert [
        TEST_GROUP_INTERNAL_REPRESENTATION_1,
        TEST_GROUP_INTERNAL_REPRESENTATION_2,
    ] == result_groups


@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_member_groups_member_not_found(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = (
        MagicMock(return_value=(None, 200, RESPONSE_HEADER_COOKIE))
    )

    result_groups = adapter.get_member_groups(user=TEST_USER, vo=TEST_VO)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_groups == []


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_attribute"
)
@patch(
    "perun.connector.perun_openapi.api.facilities_manager_api.FacilitiesManagerApi"
    ".get_assigned_resources_for_facility"
)
@patch(
    "perun.connector.perun_openapi.api.resources_manager_api.ResourcesManagerApi"
    ".get_assigned_groups"
)
@patch("perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id")
@patch("builtins.open", new_callable=mock_open)
def test_get_sp_groups_found_sp_groups(
    mock_cookie, mock_request_1, mock_request_2, mock_request_3, mock_request_4
):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    test_resources = [Resource(1, None, None, None)]

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_attribute = (
        MagicMock(
            return_value=(
                {"value": SAMPLE_SHORT_GROUP_NAME},
                200,
                RESPONSE_HEADER_COOKIE,
            )
        )
    )

    perun_openapi.api.facilities_manager_api.FacilitiesManagerApi.get_assigned_resources_for_facility = MagicMock(
        return_value=(test_resources, 200, RESPONSE_HEADER_COOKIE)
    )

    perun_openapi.api.resources_manager_api.ResourcesManagerApi.get_assigned_groups = (
        MagicMock(
            return_value=(
                [
                    TEST_GROUP_EXTERNAL_REPRESENTATION_1,
                    TEST_GROUP_EXTERNAL_REPRESENTATION_1,
                    TEST_GROUP_EXTERNAL_REPRESENTATION_2,
                    TEST_GROUP_EXTERNAL_REPRESENTATION_2,
                ],
                200,
                RESPONSE_HEADER_COOKIE,
            )
        )
    )

    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id = MagicMock(
        return_value=(TEST_VO, 200, RESPONSE_HEADER_COOKIE)
    )

    result_groups = adapter.get_sp_groups_by_facility(TEST_INTERNAL_FACILITY_1)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert [
        TEST_GROUP_INTERNAL_REPRESENTATION_1,
        TEST_GROUP_INTERNAL_REPRESENTATION_2,
    ] == result_groups


def test_get_sp_groups_no_input_facility():
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    result_groups = adapter.get_sp_groups_by_facility(None)
    assert result_groups == []


@patch(
    "perun.connector.perun_openapi.api.facilities_manager_api.FacilitiesManagerApi"
    ".get_assigned_resources_for_facility"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_sp_groups_no_resources_found(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.facilities_manager_api.FacilitiesManagerApi.get_assigned_resources_for_facility = MagicMock(
        return_value=([], 200, RESPONSE_HEADER_COOKIE)
    )

    result_groups = adapter.get_sp_groups_by_facility(TEST_INTERNAL_FACILITY_1)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_groups == []


@patch("perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id")
@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_attribute"
)
@patch(
    "perun.connector.perun_openapi.api.groups_manager_api.GroupsManagerApi.get_group_by_name"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_group_by_name(mock_cookie, mock_request_1, mock_request_2, mock_request_3):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id = MagicMock(
        return_value=(TEST_VO, 200, RESPONSE_HEADER_COOKIE)
    )

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_attribute = (
        MagicMock(
            return_value=(
                {"value": SAMPLE_SHORT_GROUP_NAME},
                200,
                RESPONSE_HEADER_COOKIE,
            )
        )
    )

    perun_openapi.api.groups_manager_api.GroupsManagerApi.get_group_by_name = MagicMock(
        return_value=(TEST_GROUP_EXTERNAL_REPRESENTATION_1, 200, RESPONSE_HEADER_COOKIE)
    )

    result = adapter.get_group_by_name(TEST_VO, "sample name")

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result == TEST_GROUP_INTERNAL_REPRESENTATION_1


@patch("perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id")
@patch(
    "perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_short_name"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_vo_correct_arguments(mock_cookie, mock_request_1, mock_request_2):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    test_vo_id = 1
    test_vo_short_name = "sample short name"
    test_vo = VO(test_vo_id, "sample name", test_vo_short_name)

    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id = MagicMock(
        return_value=(test_vo, 200, RESPONSE_HEADER_COOKIE)
    )
    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_short_name = MagicMock(
        return_value=(test_vo, 200, RESPONSE_HEADER_COOKIE)
    )

    obtained_vo_by_id = adapter.get_vo(vo_id=test_vo_id)
    obtained_vo_by_short_name = adapter.get_vo(short_name=test_vo_short_name)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert test_vo == obtained_vo_by_id
    assert test_vo == obtained_vo_by_short_name


def test_get_vo_no_arguments():
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    no_args_error_message = (
        "Neither short_name nor id was provided, "
        "please specify exactly one to find VO by."
    )

    with pytest.raises(ValueError) as error:
        _ = adapter.get_vo()

        assert str(error.value.args[0]) == no_args_error_message


def test_get_vo_too_many_arguments():
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    too_many_args_error_message = (
        "VO can be obtained either by its "
        "short_name or id, not both at the same "
        "time."
    )

    with pytest.raises(ValueError) as error:
        _ = adapter.get_vo(vo_id=1, short_name="sample short name")
        assert str(error.value.args[0]) == too_many_args_error_message


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_facility_attributes_by_names"
)
@patch("perun.connector.perun_openapi.api.searcher_api.SearcherApi" ".get_facilities")
@patch("builtins.open", new_callable=mock_open)
def test_get_facilities_by_attribute_value_correct_attribute(
    mock_cookie, mock_request_1, mock_request_2
):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.searcher_api.SearcherApi.get_facilities = MagicMock(
        return_value=(TEST_SINGLE_PERUN_FACILITY, 200, RESPONSE_HEADER_COOKIE)
    )

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_facility_attributes_by_names = MagicMock(
        return_value=(
            [
                {
                    "id": "test id",
                    "name": "test name",
                    "display_name": "test display name",
                    "type": "test type",
                    "value": TEST_INTERNAL_FACILITY_1.rp_id,
                    "namespace": "urn:perun:facility:attribute-def:def",
                    "friendlyName": "rpIdentifier",
                }
            ],
            200,
            RESPONSE_HEADER_COOKIE,
        )
    )

    attribute = {"test name": "test value"}
    result_facilities = adapter.get_facilities_by_attribute_value(attribute)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_facilities == [TEST_INTERNAL_FACILITY_1]


def test_get_facilities_by_attribute_value_empty_attribute(caplog):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    empty_attribute = {}
    wrong_number_of_attrs_error_text = (
        f"Attribute must contain exactly one "
        f"name and one value. Given attribute "
        f""
        f""
        f'contains: "{empty_attribute}".'
    )

    with caplog.at_level(logging.WARNING):
        result_facilities = adapter.get_facilities_by_attribute_value(empty_attribute)
        assert result_facilities == []
        assert wrong_number_of_attrs_error_text in caplog.text


def test_get_facilities_by_attribute_value_too_many_attributes(caplog):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    multiple_attributes = {"test name": "test value", "test name 2": "test value 2"}
    wrong_number_of_attrs_error_text = (
        f"Attribute must contain exactly one "
        f"name and one value. Given attribute "
        f""
        f""
        f'contains: "{multiple_attributes}".'
    )

    with caplog.at_level(logging.WARNING):
        result_facilities = adapter.get_facilities_by_attribute_value(
            multiple_attributes
        )
        assert result_facilities == []
        assert wrong_number_of_attrs_error_text in caplog.text


@patch(
    "perun.connector.perun_openapi.api.facilities_manager_api.FacilitiesManagerApi"
    ".get_facilities_by_attribute"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_facility_by_rp_identifier_found_facility(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.facilities_manager_api.FacilitiesManagerApi.get_facilities_by_attribute = MagicMock(
        return_value=(TEST_SINGLE_PERUN_FACILITY, 200, RESPONSE_HEADER_COOKIE)
    )

    result_facility = adapter.get_facility_by_rp_identifier(TEST_RP_IDENTIFIER_1)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_facility == TEST_INTERNAL_FACILITY_1


@patch(
    "perun.connector.perun_openapi.api.facilities_manager_api.FacilitiesManagerApi"
    ".get_facilities_by_attribute"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_facility_by_rp_identifier_no_perun_attr_found(
    mock_cookie, mock_request_1, caplog
):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.facilities_manager_api.FacilitiesManagerApi.get_facilities_by_attribute = MagicMock(
        return_value=(None, 200, RESPONSE_HEADER_COOKIE)
    )
    expected_error_message = "No facility with rpID 'test rp identifier' found."

    with pytest.raises(Exception) as error:
        _ = adapter.get_facility_by_rp_identifier(TEST_RP_IDENTIFIER_1)

        mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
        mock_cookie().write.assert_called_with(COOKIE)

        assert str(error.value.args[0]) == expected_error_message


@patch(
    "perun.connector.perun_openapi.api.facilities_manager_api.FacilitiesManagerApi"
    ".get_facilities_by_attribute"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_facility_by_rp_identifier_multiple_perun_attrs_found(
    mock_cookie, mock_request_1, caplog
):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.facilities_manager_api.FacilitiesManagerApi.get_facilities_by_attribute = MagicMock(
        return_value=(TEST_MULTIPLE_PERUN_FACILITIES, 200, RESPONSE_HEADER_COOKIE)
    )

    expected_error_message = (
        "There is more than one facility with rpID 'test rp identifier'."
    )

    with pytest.raises(Exception) as error:
        _ = adapter.get_facility_by_rp_identifier(TEST_RP_IDENTIFIER_1)

        mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
        mock_cookie().write.assert_called_with(COOKIE)

        assert str(error.value.args[0]) == expected_error_message


@patch(
    "perun.connector.perun_openapi.api.users_manager_api.UsersManagerApi"
    ".get_groups_for_facility_where_user_is_active"
)
@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_attribute"
)
@patch("perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id")
@patch("builtins.open", new_callable=mock_open)
def test_get_users_groups_on_facility_multiple_groups_found(
    mock_cookie, mock_request_1, mock_request_2, mock_request_3
):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.users_manager_api.UsersManagerApi.get_groups_for_facility_where_user_is_active = MagicMock(
        return_value=(
            [
                TEST_GROUP_EXTERNAL_REPRESENTATION_1,
                TEST_GROUP_EXTERNAL_REPRESENTATION_2,
                TEST_GROUP_EXTERNAL_REPRESENTATION_2,
                TEST_GROUP_EXTERNAL_REPRESENTATION_1,
                TEST_GROUP_EXTERNAL_REPRESENTATION_1,
            ],
            200,
            RESPONSE_HEADER_COOKIE,
        )
    )

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_attribute = (
        MagicMock(
            return_value=(
                {"value": SAMPLE_SHORT_GROUP_NAME},
                200,
                RESPONSE_HEADER_COOKIE,
            )
        )
    )

    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id = MagicMock(
        return_value=(TEST_VO, 200, RESPONSE_HEADER_COOKIE)
    )

    result_groups = adapter.get_users_groups_on_facility(
        TEST_INTERNAL_FACILITY_1, TEST_USER
    )

    for group in result_groups:
        assert isinstance(group, Group)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert [
        TEST_GROUP_INTERNAL_REPRESENTATION_1,
        TEST_GROUP_INTERNAL_REPRESENTATION_2,
    ] == result_groups


def test_get_users_groups_on_facility_no_input_facility():
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    result_groups = adapter.get_users_groups_on_facility(None, TEST_USER)

    assert result_groups == []


@patch(
    "perun.connector.perun_openapi.api.users_manager_api.UsersManagerApi"
    ".get_groups_for_facility_where_user_is_active"
)
@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_attribute"
)
@patch("perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id")
@patch("builtins.open", new_callable=mock_open)
def test_get_users_groups_on_facility_no_groups_found(
    mock_cookie, mock_request_1, mock_request_2, mock_request_3
):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.users_manager_api.UsersManagerApi.get_groups_for_facility_where_user_is_active = MagicMock(
        return_value=([], 200, RESPONSE_HEADER_COOKIE)
    )

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_attribute = (
        MagicMock(
            return_value=(
                {"value": SAMPLE_SHORT_GROUP_NAME},
                200,
                RESPONSE_HEADER_COOKIE,
            )
        )
    )

    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id = MagicMock(
        return_value=(TEST_VO, 200, RESPONSE_HEADER_COOKIE)
    )

    result_groups = adapter.get_users_groups_on_facility(
        TEST_INTERNAL_FACILITY_1, TEST_USER
    )

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_groups == []


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_facilities_by_attribute_with_attributes"
)
@patch("builtins.open", new_callable=mock_open)
def get_facilities_by_attribute_with_attributes_correct_attribute(
    mock_cookie, mock_request_1, mock_request_2
):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.searcher_api.SearcherApi.get_facilities = MagicMock(
        return_value=(TEST_SINGLE_PERUN_FACILITY, 200, RESPONSE_HEADER_COOKIE)
    )

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_facilities_by_attribute_with_attributes = MagicMock(
        return_value=(TEST_EXTERNAL_FACILITY_WITH_ATTRS_1, 200, RESPONSE_HEADER_COOKIE)
    )

    result_facilities = adapter.get_facilities_by_attribute_with_attributes(
        "test name", "test value", ["urn:perun:facility:attribute-def:def:rpIdentifier"]
    )

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_facilities == [TEST_INTERNAL_FACILITY_1]


@patch(
    "perun.connector.perun_openapi.api.users_manager_api.UsersManagerApi"
    ".get_user_by_ext_source_name_and_ext_login"
)
@patch("perun.connector.adapters.PerunRpcAdapter.PerunRpcAdapter.get_perun_user")
@patch("builtins.open", new_callable=mock_open)
def test_get_user_ext_source(mock_cookie, mock_request_1, mock_request_2):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    source_id = 1
    source_login = "john@src"
    source_name = "sample source name"
    user_ext_source_external_representation = {
        "id": source_id,
        "login": source_login,
        "ext_source": {"name": source_name},
    }

    user_ext_source_internal_representation = UserExtSource(
        source_id, source_name, source_login, TEST_USER
    )

    perun_openapi.api.users_manager_api.UsersManagerApi.get_user_ext_source_by_ext_login_and_ext_source_name = MagicMock(
        return_value=(
            user_ext_source_external_representation,
            200,
            RESPONSE_HEADER_COOKIE,
        )
    )

    PerunRpcAdapter.get_perun_user = MagicMock(return_value=TEST_USER)

    result_ext_source = adapter.get_user_ext_source(source_name, source_login)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_ext_source == user_ext_source_internal_representation


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_user_ext_source_attributes_by_names"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_user_ext_source_attributes_multiple_attributes(
    mock_cookie, mock_request_1
):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    test_attr_1_name = "perunUserAttribute_givenName"
    test_attribute_1 = {
        "id": 1,
        "name": test_attr_1_name,
        "namespace": "urn:perun:ues:attribute-def:def",
        "display_name": "sample attribute name 1",
        "type": "test type 1",
        "value": "sample attribute value 1",
        "friendlyName": "givenName",
    }

    test_attr_2_name = "perunUserAttribute_sn"
    test_attribute_2 = {
        "id": 2,
        "name": test_attr_2_name,
        "namespace": "urn:perun:ues:attribute-def:def",
        "display_name": "sample attribute name 2",
        "type": "test type 2",
        "value": "sample attribute value 2",
        "friendlyName": "sn",
    }

    attribute_names = [
        test_attribute_1["namespace"] + ":" + test_attribute_1["friendlyName"],
        test_attribute_2["namespace"] + ":" + test_attribute_2["friendlyName"],
    ]
    ext_source = UserExtSource(1, "sample name", "sample login", None)

    expected_attribute_1 = test_attribute_1.copy()

    expected_attribute_2 = test_attribute_2.copy()

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_user_ext_source_attributes_by_names = MagicMock(
        return_value=([test_attribute_1, test_attribute_2], 200, RESPONSE_HEADER_COOKIE)
    )

    result_attrs = adapter.get_user_ext_source_attributes(ext_source, attribute_names)
    expected_attrs = {
        attribute_names[0]: expected_attribute_1["value"],
        attribute_names[1]: expected_attribute_2["value"],
    }

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_attrs == expected_attrs


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_user_ext_source_attributes_by_names"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_user_ext_source_attributes_no_attributes(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_user_ext_source_attributes_by_names = MagicMock(
        return_value=([], 200, RESPONSE_HEADER_COOKIE)
    )

    ext_source = UserExtSource(1, "sample name", "sample login", None)
    result_attrs = adapter.get_user_ext_source_attributes(ext_source, [])

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_attrs == {}


@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_member_status_by_user_and_vo_valid_member(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    expected_status = "VALID"
    test_member_external_representation = {"id": 1, "status": expected_status}
    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = (
        MagicMock(
            return_value=(
                test_member_external_representation,
                200,
                RESPONSE_HEADER_COOKIE,
            )
        )
    )

    result_status = adapter.get_member_status_by_user_and_vo(TEST_USER, TEST_VO)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_status == MemberStatusEnum(expected_status)


@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_member_status_by_user_and_vo_invalid_member(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = (
        MagicMock(
            side_effect=ApiException(
                http_resp=HttpResponse('"name":"MemberNotExistsException"')
            ),
        )
    )
    expected_error_message = '"name":"MemberNotExistsException"'

    with pytest.raises(ApiException) as error:
        adapter.get_member_status_by_user_and_vo(TEST_USER, TEST_VO)
        assert str(error.value.args) == expected_error_message


@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_member_status_by_user_and_vo_invalid_status(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    invalid_status = "THIS_IS_NOT_A_VALID_STATUS"
    invalid_status_error_msg = f'"{invalid_status}" is not a valid state.'
    test_member_external_representation = {"id": 1, "status": invalid_status}

    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = (
        MagicMock(
            return_value=(
                test_member_external_representation,
                200,
                RESPONSE_HEADER_COOKIE,
            )
        )
    )

    with pytest.raises(ValueError) as error:
        _ = adapter.get_member_status_by_user_and_vo(TEST_USER, TEST_VO)

        mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
        mock_cookie().write.assert_called_with(COOKIE)

        assert str(error.value.args[0]) == invalid_status_error_msg


@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
@patch(
    "perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi"
    ".get_vo_by_short_name"
)
@patch("builtins.open", new_callable=mock_open)
def test_is_user_in_vo_valid_member(mock_cookie, mock_request_1, mock_request_2):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    valid_status_name = "VALID"
    test_valid_member_external_representation = {"id": 1, "status": valid_status_name}

    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_short_name = MagicMock(
        return_value=(TEST_VO, 200, RESPONSE_HEADER_COOKIE)
    )

    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = (
        MagicMock(
            return_value=(
                test_valid_member_external_representation,
                200,
                RESPONSE_HEADER_COOKIE,
            )
        )
    )

    user_is_valid_member = adapter.is_user_in_vo_by_short_name(
        TEST_USER, TEST_VO.short_name
    )

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert user_is_valid_member


@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
@patch(
    "perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi"
    ".get_vo_by_short_name"
)
@patch("builtins.open", new_callable=mock_open)
def test_is_user_in_vo_invalid_member(mock_cookie, mock_request_1, mock_request_2):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    invalid_status_name = "INVALID"
    test_valid_member_external_representation = {"id": 1, "status": invalid_status_name}

    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_short_name = MagicMock(
        return_value=(TEST_VO, 200, RESPONSE_HEADER_COOKIE)
    )

    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = (
        MagicMock(
            return_value=(
                test_valid_member_external_representation,
                200,
                RESPONSE_HEADER_COOKIE,
            )
        )
    )

    user_is_valid_member = adapter.is_user_in_vo_by_short_name(
        TEST_USER, TEST_VO.short_name
    )

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert not user_is_valid_member


@patch(
    "perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi"
    ".get_vo_by_short_name"
)
@patch("builtins.open", new_callable=mock_open)
def test_is_user_in_vo_non_existing_vo(mock_cookie, mock_request_1, caplog):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_short_name = MagicMock(
        side_effect=ApiException(
            http_resp=HttpResponse('"name":"VoNotExistsException"')
        ),
    )
    short_name_missing_msg = '"name":"VoNotExistsException"'

    with pytest.raises(Exception) as error:
        adapter.is_user_in_vo_by_short_name(TEST_USER, "non_existing_short_name")
        assert str(error.value.args[0]) == short_name_missing_msg


def test_is_user_in_vo_user_without_id():
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    user_without_id_msg = "User's ID is empty"
    user_without_id = User(None, "He, who shall not be named")

    with pytest.raises(ValueError) as error:
        _ = adapter.is_user_in_vo_by_short_name(user_without_id, "some short name")
        assert str(error.value.args[0]) == user_without_id_msg


def test_is_user_in_vo_no_short_name_given():
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    short_name_missing_msg = "VO short name is empty"
    missing_short_name = None

    with pytest.raises(ValueError) as error:
        _ = adapter.is_user_in_vo_by_short_name(TEST_USER, missing_short_name)
        assert str(error.value.args[0]) == short_name_missing_msg


@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_member_by_user_existing_user(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = (
        MagicMock(
            return_value=(
                TEST_MEMBER_EXTERNAL_REPRESENTATION,
                200,
                RESPONSE_HEADER_COOKIE,
            )
        )
    )

    result_member = adapter.get_member_by_user(TEST_USER, TEST_VO)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_member == TEST_MEMBER_INTERNAL_REPRESENTATION


@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_member_by_user_user_not_not_found_ex(mock_cookie, mock_request_1, caplog):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = (
        MagicMock(
            side_effect=ApiException(
                http_resp=HttpResponse('"name":"MemberNotExistsException"')
            ),
        )
    )
    short_name_missing_msg = '"name":"MemberNotExistsException"'

    with pytest.raises(Exception) as error:
        adapter.get_member_by_user(TEST_USER, TEST_VO)
        assert str(error.value.args[0]) == short_name_missing_msg


@patch(
    "perun.connector.perun_openapi.api.facilities_manager_api.FacilitiesManagerApi"
    ".get_assigned_resources_for_facility"
)
@patch(
    "perun.connector.perun_openapi.api.resources_manager_api.ResourcesManagerApi"
    ".get_assigned_groups"
)
@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_attribute"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_resource_capabilities(
    mock_cookie, mock_request_1, mock_request_2, mock_request_3
):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    group_without_resource = Group(-2, None, "", "", "", "")
    test_user_groups = [
        TEST_GROUP_INTERNAL_REPRESENTATION_1,
        TEST_GROUP_INTERNAL_REPRESENTATION_2,
        group_without_resource,
    ]

    resource_without_group = {"id": -1}
    resource_of_group_1 = {"id": test_user_groups[0].id}
    resource_of_group_2 = {"id": test_user_groups[1].id}
    test_resource_groups = [
        ([resource_of_group_1, resource_of_group_2], 200, RESPONSE_HEADER_COOKIE),
        ([resource_without_group], 200, RESPONSE_HEADER_COOKIE),
        ([], 200, RESPONSE_HEADER_COOKIE),
    ]

    test_facility_resources = [{"id": 5}, {"id": 6}, {"id": 7}]

    perun_openapi.api.facilities_manager_api.FacilitiesManagerApi.get_assigned_resources_for_facility = MagicMock(
        return_value=(test_facility_resources, 200, RESPONSE_HEADER_COOKIE)
    )

    perun_openapi.api.resources_manager_api.ResourcesManagerApi.get_assigned_groups = (
        MagicMock(side_effect=test_resource_groups)
    )

    resource_capabilities_of_groups = ["test capability 1", "test capability 2"]
    resource_capabilities_without_group = ["test capability 3"]
    absent_capabilities = None
    external_capabilities = [
        ({"value": resource_capabilities_of_groups}, 200, RESPONSE_HEADER_COOKIE),
        ({"value": resource_capabilities_without_group}, 200, RESPONSE_HEADER_COOKIE),
        ({"value": absent_capabilities}, 200, RESPONSE_HEADER_COOKIE),
    ]

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_attribute = (
        MagicMock(side_effect=external_capabilities)
    )

    expected_capabilities = ["test capability 1", "test capability 2"]
    result_capabilities = adapter.get_resource_capabilities_by_facility(
        TEST_INTERNAL_FACILITY_1, test_user_groups
    )

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert sorted(result_capabilities) == sorted(expected_capabilities)


def test_get_resource_capabilities_no_input_facility():
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    result_capabilities = adapter.get_resource_capabilities_by_facility(None, [])
    assert result_capabilities == []


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_attribute"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_facility_capabilities(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    facility_capabilities = {"value": ["test capability 1", "test capability 2"]}
    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_attribute = (
        MagicMock(return_value=(facility_capabilities, 200, RESPONSE_HEADER_COOKIE))
    )

    expected_capabilities = facility_capabilities["value"]
    result_capabilities = adapter.get_facility_capabilities_by_facility(
        TEST_INTERNAL_FACILITY_1
    )

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert sorted(result_capabilities) == sorted(expected_capabilities)


def test_get_facility_capabilities_no_input_facility():
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    result_capabilities = adapter.get_facility_capabilities_by_facility(None)
    assert result_capabilities == []


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_user_attributes_by_names"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_user_attributes_multiple_attributes(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    test_user_attributes = get_entity_specific_attributes("user")
    attr_names = [attr.get("name") for attr in test_user_attributes]

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_user_attributes_by_names = MagicMock(
        return_value=(test_user_attributes, 200, RESPONSE_HEADER_COOKIE)
    )

    expected_attributes = {
        attr.get("namespace") + ":" + attr.get("friendlyName"): attr.get("value")
        for attr in test_user_attributes
    }

    result_attributes = adapter.get_user_attributes(TEST_USER, attr_names)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_attributes == expected_attributes


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_user_attributes_by_names"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_user_attributes_no_attributes(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())

    # default attribute is loa
    default_attr = {
        "id": 1,
        "name": "perunUserAttribute_loa",
        "namespace": "urn:perun:user:attribute-def:virt",
        "display_name": "display name for loa",
        "type": "some type",
        "value": "sample value",
        "friendlyName": "loa",
    }

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_user_attributes_by_names = MagicMock(
        return_value=([default_attr], 200, RESPONSE_HEADER_COOKIE)
    )

    expected_attribute = {
        default_attr["namespace"] + ":" + default_attr["friendlyName"]: default_attr[
            "value"
        ]
    }

    result_attributes = adapter.get_user_attributes(TEST_USER, [])

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_attributes == expected_attribute


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_entityless_attributes_by_name"
)
@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_entityless_keys"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_entityless_attribute_valid_attribute(
    mock_cookie, mock_request_1, mock_request_2
):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    test_entityless_attrs = get_entity_specific_attributes("entityless")
    test_entityless_attr_name = (
        test_entityless_attrs[0].get("namespace")
        + ":"
        + test_entityless_attrs[0].get("friendlyName")
    )
    test_entityless_keys = ["entityless key 1", "entityless_key_2"]

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_entityless_attributes_by_name = MagicMock(
        return_value=(test_entityless_attrs, 200, RESPONSE_HEADER_COOKIE)
    )

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_entityless_keys = MagicMock(
        return_value=(test_entityless_keys, 200, RESPONSE_HEADER_COOKIE)
    )

    expected_attrs = {}
    for i in range(len(test_entityless_attrs)):
        expected_attrs[test_entityless_keys[i]] = test_entityless_attrs[i]["value"]
    result_attributes = adapter.get_entityless_attribute(test_entityless_attr_name)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_attributes == expected_attrs


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_entityless_attributes_by_name"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_entityless_attribute_attr_id_missing(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    test_entityless_attr_name = "perunEntitylessAttribute_orgAups"

    test_entityless_attr_without_id = {
        "id": None,
        "name": "perunEntitylessAttribute_orgAups",
        "namespace": "urn:perun:entityless:attribute-def:def:orgAups",
        "display_name": "display name for org aups",
        "type": "some type",
        "value": "sample value",
        "friendlyName": "orgAups",
    }

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_entityless_attributes_by_name = MagicMock(
        return_value=([test_entityless_attr_without_id], 200, RESPONSE_HEADER_COOKIE)
    )

    result_attributes = adapter.get_entityless_attribute(test_entityless_attr_name)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_attributes == {}


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_vo_attributes_by_names"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_vo_attributes_single_attribute(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    test_vo_attributes = get_entity_specific_attributes("vo")
    attr_names = [attr.get("name") for attr in test_vo_attributes]

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_vo_attributes_by_names = MagicMock(
        return_value=(test_vo_attributes, 200, RESPONSE_HEADER_COOKIE)
    )

    expected_attribute = {
        attr.get("namespace") + ":" + attr.get("friendlyName"): attr.get("value")
        for attr in test_vo_attributes
    }

    result_attribute = adapter.get_vo_attributes(TEST_VO, attr_names)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_attribute == expected_attribute


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_vo_attributes_by_names"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_vo_attributes_no_attributes(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    # default attribute is id
    default_attr = {
        "id": 1,
        "name": "perunVoAttribute_id",
        "namespace": "urn:perun:vo:attribute-def:core",
        "display_name": "display name for loa",
        "type": "some type",
        "value": 10,
        "friendlyName": "id",
    }

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_vo_attributes_by_names = MagicMock(
        return_value=([default_attr], 200, RESPONSE_HEADER_COOKIE)
    )

    expected_attribute = {
        default_attr["namespace"] + ":" + default_attr["friendlyName"]: default_attr[
            "value"
        ]
    }

    result_attribute = adapter.get_vo_attributes(TEST_VO, [])

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_attribute == expected_attribute


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_facility_attributes_by_names"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_facility_attribute_valid_attribute(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    test_facility_attribute = get_entity_specific_attributes("facility")[0]
    test_attr_name = (
        test_facility_attribute.get("namespace")
        + ":"
        + test_facility_attribute.get("friendlyName")
    )

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_facility_attributes_by_names = MagicMock(
        return_value=([test_facility_attribute], 200, RESPONSE_HEADER_COOKIE)
    )

    expected_attribute = test_facility_attribute.get("value")
    result_attribute = adapter.get_facility_attributes(
        TEST_INTERNAL_FACILITY_1, [test_attr_name]
    )

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result_attribute[test_attr_name] == expected_attribute


def test_get_attributes_multiple_attributes():
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    test_perun_attrs = get_entity_specific_attributes("user")
    processed_attributes = copy.deepcopy(test_perun_attrs)

    expected_attributes = {
        attr.get("namespace") + ":" + attr.get("friendlyName"): attr["value"]
        for attr in processed_attributes
    }

    result_attributes = adapter._get_attributes(test_perun_attrs)
    assert result_attributes == expected_attributes


def test_get_attributes_empty_attributes():
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    result_attributes = adapter._get_attributes([])
    assert result_attributes == {}


@patch(
    "perun.connector.perun_openapi.api.groups_manager_api.GroupsManagerApi.get_groups_where_member_is_active"
)
@patch("builtins.open", new_callable=mock_open)
def test_get_groups_where_member_is_active_not_member(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.groups_manager_api.GroupsManagerApi.get_groups_where_member_is_active = MagicMock(
        side_effect=ApiException(
            http_resp=HttpResponse('"name":"MemberNotExistsException"')
        ),
    )
    short_name_missing_msg = '"name":"MemberNotExistsException"'

    with pytest.raises(Exception) as error:
        adapter.get_groups_where_member_is_active(TEST_MEMBER_INTERNAL_REPRESENTATION)
        assert str(error.value.args[0]) == short_name_missing_msg


@patch(
    "perun.connector.perun_openapi.api.groups_manager_api.GroupsManagerApi.get_groups_where_member_is_active"
)
@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_attribute"
)
@patch("perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id")
@patch("builtins.open", new_callable=mock_open)
def test_get_groups_where_member_is_active(
    mock_cookie, mock_request_1, mock_request_2, mock_request_3
):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.groups_manager_api.GroupsManagerApi.get_groups_where_member_is_active = MagicMock(
        return_value=(
            [
                TEST_GROUP_EXTERNAL_REPRESENTATION_1,
                TEST_GROUP_EXTERNAL_REPRESENTATION_2,
            ],
            200,
            RESPONSE_HEADER_COOKIE,
        )
    )

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_attribute = (
        MagicMock(
            return_value=(
                {"value": SAMPLE_SHORT_GROUP_NAME},
                200,
                RESPONSE_HEADER_COOKIE,
            )
        )
    )

    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id = MagicMock(
        return_value=(TEST_VO, 200, RESPONSE_HEADER_COOKIE)
    )

    result = adapter.get_groups_where_member_is_active(
        TEST_MEMBER_INTERNAL_REPRESENTATION
    )

    for group in result:
        assert isinstance(group, Group)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert [
        TEST_GROUP_INTERNAL_REPRESENTATION_1,
        TEST_GROUP_INTERNAL_REPRESENTATION_2,
    ] == result


@patch(
    "perun.connector.perun_openapi.api.registrar_manager_api.RegistrarManagerApi.get_group_application_form"
)
@patch("builtins.open", new_callable=mock_open)
def test_has_registration_form_group(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.registrar_manager_api.RegistrarManagerApi.get_group_application_form = MagicMock(
        return_value=("Application", 200, RESPONSE_HEADER_COOKIE)
    )

    result = adapter.has_registration_form_group(TEST_GROUP_INTERNAL_REPRESENTATION_1)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result


@patch(
    "perun.connector.perun_openapi.api.registrar_manager_api.RegistrarManagerApi.get_group_application_form"
)
@patch("builtins.open", new_callable=mock_open)
def test_has_registration_form_group_no_form(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.registrar_manager_api.RegistrarManagerApi.get_group_application_form = MagicMock(
        return_value=([], 200, RESPONSE_HEADER_COOKIE)
    )

    result = adapter.has_registration_form_group(TEST_GROUP_INTERNAL_REPRESENTATION_1)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert not result


@patch(
    "perun.connector.perun_openapi.api.registrar_manager_api.RegistrarManagerApi.get_vo_application_form"
)
@patch("builtins.open", new_callable=mock_open)
def test_has_registration_form_vo(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.registrar_manager_api.RegistrarManagerApi.get_vo_application_form = MagicMock(
        return_value=("Application", 200, RESPONSE_HEADER_COOKIE)
    )

    result = adapter.has_registration_form_vo(TEST_VO)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert result


@patch(
    "perun.connector.perun_openapi.api.registrar_manager_api.RegistrarManagerApi.get_vo_application_form"
)
@patch("builtins.open", new_callable=mock_open)
def test_has_registration_form_vo_no_form(mock_cookie, mock_request_1):
    adapter = PerunRpcAdapter(CONFIG, ConfigStore.get_attribute_map())
    perun_openapi.api.registrar_manager_api.RegistrarManagerApi.get_vo_application_form = MagicMock(
        return_value=(None, 200, RESPONSE_HEADER_COOKIE)
    )

    result = adapter.has_registration_form_vo(TEST_VO)

    mock_cookie.assert_called_with(COOKIE_FILEPATH, "w")
    mock_cookie().write.assert_called_with(COOKIE)

    assert not result



import re
from typing import List, Optional, Union

from perun.connector import perun_openapi
from perun.connector.adapters.AdapterInterface import AdapterInterface
from perun.connector.models.Facility import Facility
from perun.connector.models.Group import Group
from perun.connector.models.Member import Member
from perun.connector.models.MemberStatusEnum import MemberStatusEnum
from perun.connector.models.Resource import Resource
from perun.connector.models.User import User
from perun.connector.models.UserExtSource import UserExtSource
from perun.connector.models.VO import VO
from perun.connector.perun_openapi import ApiClient, Configuration
from perun.connector.perun_openapi.api.attributes_manager_api import (
    AttributesManagerApi,
)
from perun.connector.perun_openapi.api.authz_resolver_api import AuthzResolverApi
from perun.connector.perun_openapi.api.facilities_manager_api import (
    FacilitiesManagerApi,
)
from perun.connector.perun_openapi.api.groups_manager_api import GroupsManagerApi
from perun.connector.perun_openapi.api.members_manager_api import MembersManagerApi
from perun.connector.perun_openapi.api.registrar_manager_api import RegistrarManagerApi
from perun.connector.perun_openapi.api.resources_manager_api import ResourcesManagerApi
from perun.connector.perun_openapi.api.searcher_api import SearcherApi
from perun.connector.perun_openapi.api.users_manager_api import UsersManagerApi
from perun.connector.perun_openapi.api.vos_manager_api import VosManagerApi
from perun.connector.perun_openapi.exceptions import ApiException
from perun.connector.perun_openapi.model.attribute import Attribute
from perun.connector.perun_openapi.model.input_get_facilities import InputGetFacilities
from perun.connector.perun_openapi.model.input_set_facility_attributes import (
    InputSetFacilityAttributes,
)
from perun.connector.perun_openapi.model.input_set_user_attributes import (
    InputSetUserAttributes,
)
from perun.connector.perun_openapi.model.input_set_user_ext_source_attributes import (
    InputSetUserExtSourceAttributes,
)
from perun.connector.utils.AttributeUtils import AttributeUtils
from perun.connector.utils.Logger import Logger


class RPCAdapterNotExistsException(Exception):
    def __init__(self, message):
        self.body = message
        super().__init__(self.body)


class PerunRpcAdapter(AdapterInterface):
    def __init__(self, config_data: dict[str, str], attrs_map):
        self._CONFIG = None
        self._logger = Logger.get_logger(self.__class__.__name__)
        self._BASIC_AUTH = "BasicAuth"
        self._BEARER_AUTH = "BearerAuth"
        self._API_KEY_AUTH = "ApiKeyAuth"

        self._set_up_openapi_config(config_data)

        if "rp_id_attribute" not in config_data:
            self._RP_ID_ATTR = "urn:perun:facility:attribute-def:def:rpIdentifier"
        else:
            self._RP_ID_ATTR = config_data["rp_id_attribute"]

        self._ATTRIBUTE_UTILS = AttributeUtils(attrs_map)
        self._COOKIE_FILEPATH = config_data.get(
            "cookie_filepath", "/tmp/perun-cookie.txt"
        )
        self.cookie = self._load_cookie_from_file(self._COOKIE_FILEPATH)

    def _set_up_openapi_config(self, config_data: dict[str, str]) -> None:
        auth_type = config_data["auth_type"]
        self._CONFIG = Configuration(host=config_data["host"])

        if auth_type == self._BASIC_AUTH:
            self._CONFIG.username = config_data["username"]
            self._CONFIG.password = config_data["password"]
        elif auth_type == self._BEARER_AUTH:
            self._CONFIG.access_token = config_data["access_token"]
        elif auth_type == self._API_KEY_AUTH:
            self._CONFIG.api_key[self._API_KEY_AUTH] = config_data["api_key"]
        else:
            exception_message = (
                f'Authentication type "{auth_type}" is not a '
                f"supported way of authentication, please "
                f'set "auth_type" to one of "'
                f'{self._BASIC_AUTH}", "{self._BEARER_AUTH}" '
                f'or "{self._API_KEY_AUTH}"'
            )
            raise ValueError(exception_message)

    def _call_api(self, callable_method, params: dict):
        params["_return_http_data_only"] = False
        data, status, headers = callable_method(**params)
        cookie = headers.get("Set-Cookie", "")
        cookie = re.match(r"PERUNSESSION=\w+;", cookie)
        if cookie:
            cookie = cookie.group()
            if cookie != self.cookie:
                self.cookie = cookie
                self._save_cookie_to_file(self._COOKIE_FILEPATH, self.cookie)
        return data

    def get_perun_user(self, idp_id: str, uids: List[str]) -> Optional[User]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            api_instance = UsersManagerApi(api_client)
            for uid in uids:
                try:
                    user = self._call_api(
                        api_instance.get_user_by_ext_source_name_and_ext_login,
                        {"ext_login": uid, "ext_source_name": idp_id},
                    )
                    name = ""
                    for user_attr in [
                        "title_before",
                        "first_name",
                        "middle_name",
                        "last_name",
                        "title_after",
                    ]:
                        if user[user_attr] is not None:
                            name += user[user_attr] + " "

                    return User(user["id"], name.strip())
                except ApiException as ex:
                    if '"name":"UserExtSourceNotExistsException"' in ex.body:
                        continue
                    raise ex
        raise RPCAdapterNotExistsException(
            f"No user with uids '{str(uids)}' for idp: {idp_id} found."
        )

    def _get_group_unique_name(
        self,
        attributes_api_instance: AttributesManagerApi,
        group_name: str,
        group_id: int,
    ) -> str:
        attr = self._call_api(
            attributes_api_instance.get_attribute,
            {
                "group": group_id,
                "attribute_name": "urn:perun:group:attribute-def:virt:voShortName",
            },
        )
        return f'{attr["value"]}:{group_name}'

    def _create_internal_representation_groups(
        self,
        input_groups: List[perun_openapi.model.group.Group],
        converted_groups: List[Group],
        attributes_api_instance: AttributesManagerApi = None,
        vo: VO = None,
    ) -> None:
        unique_ids = []
        for group in input_groups:
            if group["id"] not in unique_ids:
                group["unique_name"] = (
                    ""
                    if not attributes_api_instance
                    else self._get_group_unique_name(
                        attributes_api_instance, group["name"], group["id"]
                    )
                )
                converted_groups.append(
                    Group(
                        group["id"],
                        vo if vo else self.get_vo(vo_id=group["vo_id"]),
                        group["uuid"],
                        group["name"],
                        group["unique_name"],
                        group["description"],
                    )
                )
                unique_ids.append(group["id"])

    def get_member_groups(
        self, user: Union[User, int], vo: Union[VO, int]
    ) -> List[Group]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            members_api_instance = MembersManagerApi(api_client)
            groups_api_instance = GroupsManagerApi(api_client)
            attributes_api_instance = AttributesManagerApi(api_client)

            converted_groups = []
            if not isinstance(vo, VO):
                vo = self.get_vo(vo_id=vo)
            user_id = AdapterInterface.get_object_id(user)
            member = self._call_api(
                members_api_instance.get_member_by_user, {"vo": vo.id, "user": user_id}
            )
            member_groups = []
            if member:
                member_groups = self._call_api(
                    groups_api_instance.get_all_member_groups, {"member": member["id"]}
                )
            self._create_internal_representation_groups(
                member_groups, converted_groups, attributes_api_instance, vo
            )
            return converted_groups

    def get_sp_groups_by_facility(self, facility: Union[Facility, int]) -> List[Group]:
        if facility is None:
            return []

        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)
            facilities_api_instance = FacilitiesManagerApi(api_client)
            resources_api_instance = ResourcesManagerApi(api_client)

            facility_id = AdapterInterface.get_object_id(facility)
            resources = self._call_api(
                facilities_api_instance.get_assigned_resources_for_facility,
                {"facility": facility_id},
            )

            resources_ids = [resource.id for resource in resources]

            sp_groups = []
            for resource_id in resources_ids:
                groups = self._call_api(
                    resources_api_instance.get_assigned_groups,
                    {"resource": resource_id},
                )
                if groups:
                    vo = self.get_vo(vo_id=groups[0]["vo_id"])

                self._create_internal_representation_groups(
                    groups, sp_groups, attributes_api_instance, vo
                )
            return sp_groups

    def get_sp_groups_by_rp_id(self, rp_id: str) -> List[Group]:
        facility = self.get_facility_by_rp_identifier(rp_id)
        return self.get_sp_groups_by_facility(facility)

    def get_group_by_name(self, vo: Union[VO, int], name: str) -> Group:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)
            groups_api_instance = GroupsManagerApi(api_client)

            vo_id = AdapterInterface.get_object_id(vo)
            group = self._call_api(
                groups_api_instance.get_group_by_name, {"vo": vo_id, "name": name}
            )
            group_external_representation = [group]
            converted_group = []
            self._create_internal_representation_groups(
                group_external_representation, converted_group, attributes_api_instance
            )
            return converted_group[0]

    def get_vo(self, short_name="", vo_id=None) -> Optional[VO]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            vos_api_instance = VosManagerApi(api_client)

            if short_name and vo_id:
                raise ValueError(
                    "VO can be obtained either by its short_name or id, "
                    "not both "
                    "at the same time."
                )
            elif vo_id:
                vo_lookup_method = vos_api_instance.get_vo_by_id
                vo_lookup_attribute = {"id": vo_id}
            elif short_name:
                vo_lookup_method = vos_api_instance.get_vo_by_short_name
                vo_lookup_attribute = {"short_name": short_name}
            else:
                raise ValueError(
                    "Neither short_name nor id was provided, please specify "
                    "exactly one to find VO by."
                )
            vo = self._call_api(vo_lookup_method, vo_lookup_attribute)
            return VO(vo.id, vo.name, vo.short_name)

    def get_facility_by_rp_identifier(self, rp_identifier: str) -> Optional[Facility]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            facilities_api_instance = FacilitiesManagerApi(api_client)

            facilities = self._call_api(
                facilities_api_instance.get_facilities_by_attribute,
                {"attribute_name": self._RP_ID_ATTR, "attribute_value": rp_identifier},
            )

            if not facilities:
                raise RPCAdapterNotExistsException(
                    f"No facility with rpID '{rp_identifier}' found."
                )

            if len(facilities) > 1:
                raise RPCAdapterNotExistsException(
                    f"There is more than one facility with rpID '{rp_identifier}'."
                )
            return Facility(
                facilities[0]["id"],
                facilities[0]["name"],
                facilities[0]["description"],
                rp_identifier,
                {},
            )

    def get_users_groups_on_facility(
        self, facility: Union[Facility, int], user: Union[User, int]
    ) -> List[Group]:
        if facility is None:
            return []

        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            users_api_instance = UsersManagerApi(api_client)
            attributes_api_instance = AttributesManagerApi(api_client)

            facility_id = AdapterInterface.get_object_id(facility)
            user_id = AdapterInterface.get_object_id(user)
            users_groups_on_facility = self._call_api(
                users_api_instance.get_groups_for_facility_where_user_is_active,
                {"user": user_id, "facility": facility_id},
            )
            converted_groups = []
            self._create_internal_representation_groups(
                users_groups_on_facility, converted_groups, attributes_api_instance
            )
            return converted_groups

    def get_users_groups_on_facility_by_rp_id(
        self, rp_identifier: str, user: Union[User, int]
    ) -> List[Group]:
        facility = self.get_facility_by_rp_identifier(rp_identifier)
        return self.get_users_groups_on_facility(facility, user)

    def _get_rp_id(self, facility: Union[Facility, int]) -> str:
        return self.get_facility_attributes(facility, [self._RP_ID_ATTR]).get(
            self._RP_ID_ATTR
        )

    def get_facilities_by_attribute_value(
        self, attribute: dict[str, str]
    ) -> List[Facility]:
        if len(attribute) != 1:
            self._logger.warning(
                f"Attribute must contain exactly one name and one value. "
                f'Given attribute contains: "{attribute}".'
            )
            return []

        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            searcher_api = SearcherApi(api_client)

            attribute_to_match_in_facilities = InputGetFacilities(attribute)
            perun_facilities = self._call_api(
                searcher_api.get_facilities,
                {"input_get_facilities": attribute_to_match_in_facilities},
            )

            facilities = []
            for perun_facility in perun_facilities:
                facility = Facility(
                    perun_facility["id"],
                    perun_facility["name"],
                    perun_facility["description"],
                    "",
                    {},
                )
                facility.rp_id = self._get_rp_id(facility)
                facilities.append(facility)

        return facilities

    def get_facility_attributes(
        self, facility: Union[Facility, int], attr_names: List[str]
    ) -> dict[str, Union[str, Optional[int], bool, List[str], dict[str, str]]]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)

            facility_id = AdapterInterface.get_object_id(facility)
            perun_attrs = self._call_api(
                attributes_api_instance.get_facility_attributes_by_names,
                {"facility": facility_id, "attr_names": attr_names},
            )
            return self._get_attributes(perun_attrs)

    def get_user_ext_source(
        self, ext_source_name: str, ext_source_login: str
    ) -> UserExtSource:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            users_api_instance = UsersManagerApi(api_client)

            user_ext_source_perun = self._call_api(
                users_api_instance.get_user_ext_source_by_ext_login_and_ext_source_name,
                {
                    "ext_source_name": ext_source_name,
                    "ext_source_login": ext_source_login,
                },
            )

            ext_source_id = user_ext_source_perun["id"]
            login = user_ext_source_perun["login"]

            ext_source_details = user_ext_source_perun["ext_source"]
            name = ext_source_details["name"]

            user = self.get_perun_user(ext_source_name, [ext_source_login])

            return UserExtSource(ext_source_id, name, login, user)

    def get_user_ext_source_by_unique_attribute_value(
        self,
        attr_name: str,
        attr_value: Union[str, int, bool, List[str], dict[str, str]],
    ) -> UserExtSource:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            api_function = UsersManagerApi(
                api_client
            ).get_user_ext_source_by_unique_attribute_value_and_attribute_name

            try:
                user_ext_source_perun = self._call_api(
                    api_function,
                    {"attribute_name": attr_name, "attribute_value": attr_value},
                )
            except ApiException as ex:
                if '"name":"UserExtSourceNotExistsException"' in ex.body:
                    raise RPCAdapterNotExistsException(
                        f"No user ext source for attr name: '{attr_value}' value: "
                        f"{str(attr_value)} found."
                    )

            ext_source_id = user_ext_source_perun["id"]
            login = user_ext_source_perun["login"]

            ext_source_details = user_ext_source_perun["ext_source"]
            name = ext_source_details["name"]

            user = self.get_perun_user(name, [login])

            return UserExtSource(ext_source_id, name, login, user)

    def update_user_ext_source_last_access(
        self, user_ext_source: Union[UserExtSource, int]
    ) -> None:
        user_ext_source_id = AdapterInterface.get_object_id(user_ext_source)

        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            users_api_instance = UsersManagerApi(api_client)

            self._call_api(
                users_api_instance.update_user_ext_source_last_access,
                {"user_ext_source": user_ext_source_id},
            )

    def get_user_ext_source_attributes(
        self, user_ext_source: Union[UserExtSource, int], attr_names: List[str]
    ) -> dict[str, Union[str, Optional[int], bool, List[str], dict[str, str]]]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)

            user_ext_source_id = AdapterInterface.get_object_id(user_ext_source)
            perun_attrs = self._call_api(
                attributes_api_instance.get_user_ext_source_attributes_by_names,
                {"user_ext_source": user_ext_source_id, "attr_names": attr_names},
            )
            return self._get_attributes(perun_attrs)

    def set_user_ext_source_attributes(
        self,
        user_ext_source: Union[UserExtSource, int],
        attributes: dict[
            str, Union[str, Optional[int], bool, List[str], dict[str, str]]
        ],
    ) -> None:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)
            user_ext_source_id = AdapterInterface.get_object_id(user_ext_source)
            openapi_attributes = self.internal_to_open_api_attrs(attributes)
            input_arg = InputSetUserExtSourceAttributes(
                user_ext_source_id, openapi_attributes
            )
            self._call_api(
                attributes_api_instance.set_user_ext_source_attributes,
                {"input_set_user_ext_source_attributes": input_arg},
            )

    def set_user_attributes(
        self,
        user: Union[User, int],
        attributes: dict[
            str, Union[str, Optional[int], bool, List[str], dict[str, str]]
        ],
    ) -> None:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)
            user_id = AdapterInterface.get_object_id(user)
            openapi_attributes = self.internal_to_open_api_attrs(attributes)
            input_arg = InputSetUserAttributes(user_id, openapi_attributes)
            self._call_api(
                attributes_api_instance.set_user_attributes,
                {"input_set_user_attributes": input_arg},
            )

    def internal_to_open_api_attrs(
        self,
        attributes: dict[
            str, Union[str, Optional[int], bool, List[str], dict[str, str]]
        ],
    ) -> List[Attribute]:
        attrs_cfg = self._ATTRIBUTE_UTILS.get_specific_attrs_config_dict(
            list(attributes.keys())
        )
        openapi_attributes = []
        for key, value in attrs_cfg.items():
            name_parts = key.split(":")
            namespace = ":".join(name_parts[:5])
            friendly_name = ":".join(name_parts[5:])
            openapi_attributes.append(
                Attribute(
                    id=value["id"],
                    bean_name="Attribute",
                    namespace=namespace,
                    friendly_name=friendly_name,
                    type=value["type"],
                    value=attributes[key],
                )
            )
        return openapi_attributes

    def get_member_status_by_user_and_vo(
        self, user: Union[User, int], vo: Union[VO, int]
    ) -> Optional[str]:
        member = self.get_member_by_user(user, vo)

        if member is not None:
            return member.status

        return None

    def is_user_in_vo_by_short_name(
        self, user: Union[User, int], vo_short_name: str
    ) -> bool:
        user_id = AdapterInterface.get_object_id(user)
        if not user_id:
            raise ValueError("User's ID is empty")

        if not vo_short_name:
            raise ValueError("VO short name is empty")

        vo_of_user = self.get_vo(short_name=vo_short_name)

        user_status = self.get_member_status_by_user_and_vo(user, vo_of_user)
        valid_status = MemberStatusEnum.VALID

        return user_status == valid_status

    def get_member_by_user(
        self, user: Union[User, int], vo: Union[VO, int]
    ) -> Optional[Member]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            members_api_instance = MembersManagerApi(api_client)

            user_id = AdapterInterface.get_object_id(user)
            vo_id = AdapterInterface.get_object_id(vo)

            member = self._call_api(
                members_api_instance.get_member_by_user, {"vo": vo_id, "user": user_id}
            )
            return Member(member["id"], vo, member["status"])

    def get_resource_capabilities_by_facility(
        self, facility: Union[Facility, int], user_groups: List[Union[Group, int]]
    ) -> List[str]:
        capabilities = []
        if facility is None:
            return capabilities

        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            facilities_api_instance = FacilitiesManagerApi(api_client)
            resources_api_instance = ResourcesManagerApi(api_client)
            attributes_api_instance = AttributesManagerApi(api_client)

            facility_id = AdapterInterface.get_object_id(facility)
            resources = self._call_api(
                facilities_api_instance.get_assigned_resources_for_facility,
                {"facility": facility_id},
            )
            user_groups_ids = [
                AdapterInterface.get_object_id(user_group) for user_group in user_groups
            ]
            for resource in resources:
                resource_groups = self._call_api(
                    resources_api_instance.get_assigned_groups,
                    {"resource": resource["id"]},
                )

                attr_name = "urn:perun:resource:attribute-def:def:capabilities"
                resource_capabilities = self._call_api(
                    attributes_api_instance.get_attribute,
                    {
                        "resource": resource["id"],
                        "attribute_name": attr_name,
                    },
                )["value"]

                if resource_capabilities is None:
                    continue

                for resource_group in resource_groups:
                    if resource_group["id"] in user_groups_ids:
                        capabilities.extend(resource_capabilities)
                        break
        return capabilities

    def get_resource_capabilities_by_rp_id(
        self, rp_identifier: str, user_groups: List[Union[Group, int]]
    ) -> List[str]:
        facility = self.get_facility_by_rp_identifier(rp_identifier)
        return self.get_resource_capabilities_by_facility(facility, user_groups)

    def get_facility_capabilities_by_facility(
        self, facility: Union[Facility, int]
    ) -> List[str]:
        if facility is None:
            return []

        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)

            facility_id = AdapterInterface.get_object_id(facility)

            attr_name = "urn:perun:facility:attribute-def:def:capabilities"
            facility_capabilities = self._call_api(
                attributes_api_instance.get_attribute,
                {
                    "facility": facility_id,
                    "attribute_name": attr_name,
                },
            )["value"]

            return facility_capabilities

    def get_facility_capabilities_by_rp_id(self, rp_identifier: str) -> List[str]:
        facility = self.get_facility_by_rp_identifier(rp_identifier)
        return self.get_facility_capabilities_by_facility(facility)

    def get_resource_attributes(
        self, resource: Union[Resource, int], attr_names: List[str]
    ) -> dict[str, Union[str, Optional[int], bool, List[str], dict[str, str]]]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)

            resource_id = AdapterInterface.get_object_id(resource)
            perun_attrs = self._call_api(
                attributes_api_instance.get_resource_attributes_by_names,
                {"resource": resource_id, "attr_names": attr_names},
            )
            return self._get_attributes(perun_attrs)

    def get_user_attributes(
        self, user: Union[User, int], attr_names: List[str]
    ) -> dict[str, Union[str, Optional[int], bool, List[str], dict[str, str]]]:
        default_attribute_name = "urn:perun:user:attribute-def:virt:loa"
        if not attr_names:
            attr_names.append(default_attribute_name)

        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)
            user_id = AdapterInterface.get_object_id(user)

            perun_attrs = self._call_api(
                attributes_api_instance.get_user_attributes_by_names,
                {"user": user_id, "attr_names": attr_names},
            )

            return self._get_attributes(perun_attrs)

    def get_entityless_attribute(
        self, attr_name: str
    ) -> Union[str, Optional[int], bool, List[str], dict[str, str]]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)

            attributes = {}
            perun_attr_values = self._call_api(
                attributes_api_instance.get_entityless_attributes_by_name,
                {"attr_name": attr_name},
            )

            attr_id = perun_attr_values[0].get("id")
            if attr_id is None:
                return attributes

            perun_attr_keys = self._call_api(
                attributes_api_instance.get_entityless_keys,
                {"attribute_definition": attr_id},
            )

            for i in range(len(perun_attr_values)):
                attributes[perun_attr_keys[i]] = perun_attr_values[i]["value"]
            return attributes

    def get_vo_attributes(
        self, vo: Union[VO, int], attr_names: List[str]
    ) -> dict[str, Union[str, Optional[int], bool, List[str], dict[str, str]]]:
        default_attribute_name = "urn:perun:vo:attribute-def:core:id"
        if not attr_names:
            attr_names.append(default_attribute_name)

        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)

            vo_id = AdapterInterface.get_object_id(vo)

            perun_attrs = self._call_api(
                attributes_api_instance.get_vo_attributes_by_names,
                {"vo": vo_id, "attr_names": attr_names},
            )

            return self._get_attributes(perun_attrs)

    def _get_attributes(
        self,
        perun_attrs: List[dict[str, Union[str, int, bool, list[str], dict[str, str]]]],
    ) -> dict[str, Union[str, int, bool, list[str], dict[str, str]]]:
        attributes = {}
        for perun_attr in perun_attrs:
            perun_attr_name = perun_attr["namespace"] + ":" + perun_attr["friendlyName"]

            # for some reason open_api map int attributes as float
            if perun_attr["type"] == AttributeUtils.PERUN_INT:
                attributes[perun_attr_name] = int(perun_attr["value"])
            else:
                attributes[perun_attr_name] = perun_attr["value"]

        return attributes

    def get_groups_where_member_is_active(
        self, member: Union[Member, int]
    ) -> List[Group]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            member_id = self.get_object_id(member)
            internal_groups = []
            groups_api_instance = GroupsManagerApi(api_client)
            perun_groups = self._call_api(
                groups_api_instance.get_groups_where_member_is_active,
                {"member": member_id},
            )
            if isinstance(member, Member):
                vo = member.vo
            elif perun_groups:
                vo = self.get_vo(vo_id=perun_groups[0]["vo_id"])
            attributes_api_instance = AttributesManagerApi(api_client)
            self._create_internal_representation_groups(
                perun_groups, internal_groups, attributes_api_instance, vo
            )
            return internal_groups

    def get_groups_where_user_as_member_is_active(
        self, user: Union[User, int], vo: Union[VO, int]
    ) -> List[Group]:
        member = self.get_member_by_user(user, vo)
        return self.get_groups_where_member_is_active(member)

    def has_registration_form_group(self, group: Union[Group, int]) -> bool:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            group_id = self.get_object_id(group)
            registrar_api_instance = RegistrarManagerApi(api_client)
            if not self._call_api(
                registrar_api_instance.get_group_application_form, {"group": group_id}
            ):
                return False
            return True

    def has_registration_form_vo(self, vo: Union[VO, int]) -> bool:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            vo_id = self.get_object_id(vo)
            registrar_api_instance = RegistrarManagerApi(api_client)
            if not self._call_api(
                registrar_api_instance.get_vo_application_form, {"vo": vo_id}
            ):
                return False
            return True

    def has_registration_form_by_vo_short_name(self, vo_short_name: str) -> bool:
        vo = self.get_vo(vo_short_name)
        return self.has_registration_form_vo(vo)

    def create_facility(self, name: str, description="") -> Facility:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            facilities_api_instance = FacilitiesManagerApi(api_client)
            if description:
                return self._call_api(
                    facilities_api_instance.create_facility,
                    {"name": name, "description": description},
                )
            return self._call_api(
                facilities_api_instance.create_facility, {"name": name}
            )

    def set_facility_attributes(
        self,
        facility: Union[Facility, int],
        attributes: dict[
            str, Union[str, Optional[int], bool, List[str], dict[str, str]]
        ],
    ) -> None:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)
            facility_id = AdapterInterface.get_object_id(facility)
            openapi_attributes = self.internal_to_open_api_attrs(attributes)

            self._call_api(
                attributes_api_instance.set_facility_attributes,
                {
                    "input_set_facility_attributes": InputSetFacilityAttributes(
                        facility=facility_id, attributes=openapi_attributes
                    )
                },
            )

    def get_attributes_definition(self) -> List[dict[str, Union[str, int, bool]]]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)
            return self._call_api(
                attributes_api_instance.get_all_attribute_definitions, {}
            )

    def is_user_perun_admin(self, user: Union[User, int]) -> bool:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            authz_api_instance = AuthzResolverApi(api_client)
            if "perunadmin" in self._call_api(
                authz_api_instance.get_user_role_names,
                {"user": AdapterInterface.get_object_id(user)},
            ):
                return True
            return False

    def get_vos_where_user_is_admin(
        self, user: Union[User, int], check_perun_admin=True
    ) -> List[VO]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            if check_perun_admin and self.is_user_perun_admin(user):
                return self.get_all_vos()
            users_api_instance = UsersManagerApi(api_client)
            vos = self._call_api(
                users_api_instance.get_vos_where_user_is_admin,
                {"user": AdapterInterface.get_object_id(user)},
            )
            return [VO(vo.id, vo.name, vo.short_name) for vo in vos]

    def get_groups_where_user_is_admin(
        self,
        user: Union[User, int],
        check_perun_admin=True,
        fill_group_unique_name=True,
        fetch_related_vos=True,
    ) -> List[Group]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            if check_perun_admin and self.is_user_perun_admin(user):
                return self.get_all_groups(fetch_related_vos, fill_group_unique_name)
            else:
                users_api_instance = UsersManagerApi(api_client)
                groups = self._call_api(
                    users_api_instance.get_groups_where_user_is_admin,
                    {"user": AdapterInterface.get_object_id(user)},
                )
            return self._api_to_internal_groups(
                api_client, groups, fetch_related_vos, fill_group_unique_name
            )

    def get_facilities_where_user_is_admin(
        self, user: Union[User, int], check_perun_admin=True, fill_facility_rp_id=True
    ) -> List[Facility]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            if check_perun_admin and self.is_user_perun_admin(user):
                return self.get_all_facilities(fill_facility_rp_id)
            facilities_api_instance = FacilitiesManagerApi(api_client)
            facilities = self._call_api(
                facilities_api_instance.get_facilities_where_user_is_admin,
                {"user": AdapterInterface.get_object_id(user)},
            )
            return [
                Facility(
                    facility.id,
                    facility.name,
                    facility.description,
                    self._get_rp_id(facility.id) if fill_facility_rp_id else "",
                    {},
                )
                for facility in facilities
            ]

    def get_resources_where_user_is_admin(
        self,
        user: Union[User, int],
        check_perun_admin=True,
        fetch_related_vos=True,
        fetch_related_facilities=True,
        fill_facility_rp_id=True,
    ) -> List[Resource]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            resources_api_instance = ResourcesManagerApi(api_client)

            facilities = (
                self.get_all_facilities(fill_facility_rp_id)
                if fetch_related_facilities
                else []
            )
            vos = self.get_all_vos() if fetch_related_vos else []

            if check_perun_admin and self.is_user_perun_admin(user):
                resources = self._call_api(resources_api_instance.get_all_resources, {})
            else:
                resources = self._call_api(
                    resources_api_instance.get_all_resources_where_user_is_admin,
                    {"user": AdapterInterface.get_object_id(user)},
                )
            return [
                Resource(
                    resource.id,
                    self._find_vo_in_list_by_id(vos, resource.vo_id),
                    self._find_facility_in_list_by_id(facilities, resource.facility_id),
                    resource.name,
                )
                for resource in resources
            ]

    def get_all_vos(self) -> List[VO]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            vos_api_instance = VosManagerApi(api_client)
            vos = self._call_api(vos_api_instance.get_all_vos, {})
            return [VO(vo.id, vo.name, vo.short_name) for vo in vos]

    def get_all_facilities(self, fill_facility_rp_id=True) -> List[Facility]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            facilities_api_instance = FacilitiesManagerApi(api_client)
            facilities = self._call_api(facilities_api_instance.get_all_facilities, {})
            return [
                Facility(
                    facility.id,
                    facility.name,
                    facility.description,
                    "" if not fill_facility_rp_id else self._get_rp_id(facility.id),
                    {},
                )
                for facility in facilities
            ]

    def get_all_groups(
        self, fetch_related_vos=True, fill_group_unique_name=True
    ) -> List[Group]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            groups_api_instance = GroupsManagerApi(api_client)
            groups = self._call_api(groups_api_instance.get_all_groups_from_all_vos, {})
            return self._api_to_internal_groups(
                api_client, groups, fetch_related_vos, fill_group_unique_name
            )

    def get_all_resources(
        self,
        fetch_related_vos=True,
        fetch_related_facilities=True,
        fill_facility_rp_id=True,
    ) -> List[Resource]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            resources_api_instance = ResourcesManagerApi(api_client)
            resources = self._call_api(resources_api_instance.get_all_resources, {})
            return self._api_to_internal_resource(
                resources,
                fetch_related_vos,
                fetch_related_facilities,
                fill_facility_rp_id,
            )

    def get_groups_where_user_is_active_resource(
        self,
        user: Union[User, int],
        resource: Union[Resource, int],
        map_unique_name=True,
    ) -> List[Group]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            vo = None
            if isinstance(resource, Resource):
                vo = resource.vo
                resource_id = resource.id
            else:
                resource_id = resource

            if not vo:
                perun_vo = self._call_api(
                    ResourcesManagerApi(api_client).get_vo,
                    {"resource": resource_id},
                )
                vo = VO(perun_vo.id, perun_vo.name, perun_vo.short_name)

            result = []
            perun_groups = UsersManagerApi(
                api_client
            ).get_groups_for_resource_where_user_is_active(
                self.get_object_id(user), resource_id
            )
            self._create_internal_representation_groups(
                perun_groups,
                result,
                AttributesManagerApi(api_client) if map_unique_name else "",
                vo=vo,
            )
            return result

    def get_facility_by_id(self, facility_id: int) -> Facility:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            facility = self._call_api(
                FacilitiesManagerApi(api_client).get_facility_by_id, {"id": facility_id}
            )
            return Facility(
                facility_id,
                facility.name,
                facility.description,
                self._get_rp_id(facility_id),
                {},
            )

    def get_resources_for_facility(
        self, facility: Union[Facility, int], map_vo=True
    ) -> List[Resource]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            if isinstance(facility, Facility):
                facility_id = facility.id
            else:
                facility_id = facility
                facility = self.get_facility_by_id(facility)

            resources = self._call_api(
                FacilitiesManagerApi(api_client).get_assigned_resources_for_facility,
                {"facility": facility_id},
            )
            return [
                Resource(
                    resource.id,
                    self.get_vo(vo_id=resource.vo_id)
                    if map_vo
                    else VO(resource.vo_id, "", ""),
                    facility,
                    resource.name,
                )
                for resource in resources
            ]

    def get_groups_for_resource(
        self, resource: Union[Resource, int], map_unique_name=True
    ) -> List[Group]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            if isinstance(resource, Resource):
                resource_id = resource.id
            else:
                resource_id = resource
                resource = self.get_resource_by_id(resource_id)
            groups = self._call_api(
                ResourcesManagerApi(api_client).get_assigned_groups,
                {"resource": resource_id},
            )
            result = []
            self._create_internal_representation_groups(
                groups,
                result,
                AttributesManagerApi(api_client) if map_unique_name else "",
                vo=resource.vo,
            )
            return result

    def get_resource_by_id(
        self, resource_id: int, map_vo=True, map_facility=True
    ) -> Resource:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            resource = self._call_api(
                ResourcesManagerApi(api_client).get_resource_by_id, {"id": resource_id}
            )
            facility = (
                self.get_facility_by_id(resource.facility_id)
                if map_facility
                else Facility(resource.facility_id, "", "", "", {})
            )
            vo = (
                self.get_vo(vo_id=resource.vo_id)
                if map_vo
                else VO(resource.vo_id, "", "")
            )
            return Resource(resource.id, vo, facility, resource.name)

    def get_all_parent_groups_for_group(
        self, group: Union[Group, int], map_unique_name=True
    ) -> List[Group]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            perun_parent_groups = []
            perun_group = self._call_api(
                GroupsManagerApi(api_client).get_group_by_id,
                {"id": self.get_object_id(group)},
            )
            if isinstance(group, Group):
                vo = group.vo
            else:
                vo = self.get_vo(vo_id=perun_group.vo_id)
            if not perun_group.parent_group_id:
                return perun_parent_groups

            while perun_group.parent_group_id:
                perun_group = self._call_api(
                    GroupsManagerApi(api_client).get_parent_group,
                    {"group": perun_group.id},
                )
                perun_parent_groups.append(perun_group)
            result = []
            self._create_internal_representation_groups(
                perun_parent_groups,
                result,
                AttributesManagerApi(api_client) if map_unique_name else "",
                vo=vo,
            )
            return result

    def get_facilities_by_attribute_with_attributes(
        self,
        attr_name: str,
        attr_value: Union[str, int, bool, List[str], dict[str, str]],
        attr_names: List[str],
    ) -> List[Facility]:
        with ApiClient(self._CONFIG, cookie=self.cookie) as api_client:
            if self._RP_ID_ATTR not in attr_name:
                attr_names.append(self._RP_ID_ATTR)

            perun_facilities_attributes = self._call_api(
                FacilitiesManagerApi(
                    api_client
                ).get_facilities_by_attribute_with_attributes,
                {
                    "attribute_name": attr_name,
                    "attribute_value": attr_value,
                    "attr_names": attr_names,
                },
            )

            facilities = []
            for facility_attrs in perun_facilities_attributes:
                perun_facility = facility_attrs["facility"]
                attrs = self._get_attributes(facility_attrs["attributes"])
                facilities.append(
                    Facility(
                        perun_facility.id,
                        perun_facility.name,
                        perun_facility.description,
                        attrs[self._RP_ID_ATTR],
                        attrs,
                    )
                )
            return facilities

    def _save_cookie_to_file(self, filepath, cookie):
        with open(filepath, "w") as file:
            try:
                file.write(cookie)
            except OSError as e:
                self._logger(
                    f"Cannot write cookie to file: {filepath}, error: {e.strerror}"
                )

    def _api_to_internal_groups(
        self,
        api_client,
        api_groups,
        fetch_related_vos=True,
        fill_group_unique_name=True,
    ) -> List[Group]:
        vos = self.get_all_vos() if fetch_related_vos else []
        attributes_api_instance = AttributesManagerApi(api_client)
        return [
            Group(
                group.id,
                self._find_vo_in_list_by_id(vos, group.vo_id),
                group.uuid,
                group.name,
                ""
                if not fill_group_unique_name
                else self._get_group_unique_name(
                    attributes_api_instance, group.name, group.id
                ),
                group.description,
            )
            for group in api_groups
        ]

    def _api_to_internal_resource(
        self,
        api_resources,
        fetch_related_vos=True,
        fetch_related_facilities=True,
        fill_facility_rp_id=True,
    ) -> List[Resource]:
        facilities = (
            self.get_all_facilities(fill_facility_rp_id)
            if fetch_related_facilities
            else []
        )
        vos = self.get_all_vos() if fetch_related_vos else []
        return [
            Resource(
                resource.id,
                self._find_vo_in_list_by_id(vos, resource.vo_id),
                self._find_facility_in_list_by_id(facilities, resource.facility_id),
                resource.name,
            )
            for resource in api_resources
        ]

    @staticmethod
    def _find_vo_in_list_by_id(vo_list, vo_id):
        for vo in vo_list:
            if vo.id == vo_id:
                return vo
        return None

    @staticmethod
    def _find_facility_in_list_by_id(facility_list, facility_id):
        for facility in facility_list:
            if facility.id == facility_id:
                return facility
        return None

    @staticmethod
    def _load_cookie_from_file(filepath):
        try:
            with open(filepath, "r") as f:
                return f.read()
        except OSError:
            return None



"""
    Perun RPC API

    Perun Remote Procedure Calls Application Programming Interface  # noqa: E501

    The version of the OpenAPI document: 0.0.0
    Contact: perun@cesnet.cz
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from perun.connector.perun_openapi.model_utils import (  # noqa: F401
    ApiTypeError,
    ModelComposed,
    ModelNormal,
    ModelSimple,
    cached_property,
    change_keys_js_to_python,
    convert_js_args_to_python_args,
    date,
    datetime,
    file_type,
    none_type,
    validate_get_composed_info,
    OpenApiModel,
)
from perun.connector.perun_openapi.exceptions import ApiAttributeError


def lazy_import():
    from perun.connector.perun_openapi.model.attribute import Attribute
    from perun.connector.perun_openapi.model.attribute_definition import (
        AttributeDefinition,
    )
    from perun.connector.perun_openapi.model.auditable import Auditable
    from perun.connector.perun_openapi.model.author import Author
    from perun.connector.perun_openapi.model.authorship import Authorship
    from perun.connector.perun_openapi.model.ban import Ban
    from perun.connector.perun_openapi.model.ban_on_facility import BanOnFacility
    from perun.connector.perun_openapi.model.ban_on_resource import BanOnResource
    from perun.connector.perun_openapi.model.ban_on_vo import BanOnVo
    from perun.connector.perun_openapi.model.candidate import Candidate
    from perun.connector.perun_openapi.model.category import Category
    from perun.connector.perun_openapi.model.consent import Consent
    from perun.connector.perun_openapi.model.consent_hub import ConsentHub
    from perun.connector.perun_openapi.model.destination import Destination
    from perun.connector.perun_openapi.model.ext_source import ExtSource
    from perun.connector.perun_openapi.model.facility import Facility
    from perun.connector.perun_openapi.model.group import Group
    from perun.connector.perun_openapi.model.host import Host
    from perun.connector.perun_openapi.model.member import Member
    from perun.connector.perun_openapi.model.owner import Owner
    from perun.connector.perun_openapi.model.publication import Publication
    from perun.connector.perun_openapi.model.publication_for_gui import (
        PublicationForGUI,
    )
    from perun.connector.perun_openapi.model.publication_system import PublicationSystem
    from perun.connector.perun_openapi.model.resource import Resource
    from perun.connector.perun_openapi.model.resource_tag import ResourceTag
    from perun.connector.perun_openapi.model.rich_destination import RichDestination
    from perun.connector.perun_openapi.model.rich_facility import RichFacility
    from perun.connector.perun_openapi.model.rich_group import RichGroup
    from perun.connector.perun_openapi.model.rich_member import RichMember
    from perun.connector.perun_openapi.model.rich_resource import RichResource
    from perun.connector.perun_openapi.model.rich_user import RichUser
    from perun.connector.perun_openapi.model.security_team import SecurityTeam
    from perun.connector.perun_openapi.model.service import Service
    from perun.connector.perun_openapi.model.service_for_gui import ServiceForGUI
    from perun.connector.perun_openapi.model.services_package import ServicesPackage
    from perun.connector.perun_openapi.model.task_result import TaskResult
    from perun.connector.perun_openapi.model.thanks import Thanks
    from perun.connector.perun_openapi.model.thanks_for_gui import ThanksForGUI
    from perun.connector.perun_openapi.model.user import User
    from perun.connector.perun_openapi.model.user_ext_source import UserExtSource
    from perun.connector.perun_openapi.model.vo import Vo

    globals()["Attribute"] = Attribute
    globals()["AttributeDefinition"] = AttributeDefinition
    globals()["Auditable"] = Auditable
    globals()["Author"] = Author
    globals()["Authorship"] = Authorship
    globals()["Ban"] = Ban
    globals()["BanOnFacility"] = BanOnFacility
    globals()["BanOnResource"] = BanOnResource
    globals()["BanOnVo"] = BanOnVo
    globals()["Candidate"] = Candidate
    globals()["Category"] = Category
    globals()["Consent"] = Consent
    globals()["ConsentHub"] = ConsentHub
    globals()["Destination"] = Destination
    globals()["ExtSource"] = ExtSource
    globals()["Facility"] = Facility
    globals()["Group"] = Group
    globals()["Host"] = Host
    globals()["Member"] = Member
    globals()["Owner"] = Owner
    globals()["Publication"] = Publication
    globals()["PublicationForGUI"] = PublicationForGUI
    globals()["PublicationSystem"] = PublicationSystem
    globals()["Resource"] = Resource
    globals()["ResourceTag"] = ResourceTag
    globals()["RichDestination"] = RichDestination
    globals()["RichFacility"] = RichFacility
    globals()["RichGroup"] = RichGroup
    globals()["RichMember"] = RichMember
    globals()["RichResource"] = RichResource
    globals()["RichUser"] = RichUser
    globals()["SecurityTeam"] = SecurityTeam
    globals()["Service"] = Service
    globals()["ServiceForGUI"] = ServiceForGUI
    globals()["ServicesPackage"] = ServicesPackage
    globals()["TaskResult"] = TaskResult
    globals()["Thanks"] = Thanks
    globals()["ThanksForGUI"] = ThanksForGUI
    globals()["User"] = User
    globals()["UserExtSource"] = UserExtSource
    globals()["Vo"] = Vo


class PerunBean(ModelNormal):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Attributes:
      allowed_values (dict): The key is the tuple path to the attribute
          and the for var_name this is (var_name,). The value is a dict
          with a capitalized key describing the allowed value and an allowed
          value. These dicts store the allowed enum values.
      attribute_map (dict): The key is attribute name
          and the value is json key in definition.
      discriminator_value_class_map (dict): A dict to go from the discriminator
          variable value to the discriminator class name.
      validations (dict): The key is the tuple path to the attribute
          and the for var_name this is (var_name,). The value is a dict
          that stores validations for max_length, min_length, max_items,
          min_items, exclusive_maximum, inclusive_maximum, exclusive_minimum,
          inclusive_minimum, and regex.
      additional_properties_type (tuple): A tuple of classes accepted
          as additional properties values.
    """

    allowed_values = {}

    validations = {}

    @cached_property
    def additional_properties_type():
        """
        This must be a method because a model may have properties that are
        of type self, this must run after the class is loaded
        """
        lazy_import()
        return (
            bool,
            date,
            datetime,
            dict,
            float,
            int,
            list,
            str,
            none_type,
        )  # noqa: E501

    _nullable = False

    @cached_property
    def openapi_types():
        """
        This must be a method because a model may have properties that are
        of type self, this must run after the class is loaded

        Returns
            openapi_types (dict): The key is attribute name
                and the value is attribute type.
        """
        lazy_import()
        return {
            "id": (int,),  # noqa: E501
            "bean_name": (str,),  # noqa: E501
        }

    @cached_property
    def discriminator():
        lazy_import()
        val = {
            "Attribute": Attribute,
            "AttributeDefinition": AttributeDefinition,
            "Auditable": Auditable,
            "Author": Author,
            "Authorship": Authorship,
            "Ban": Ban,
            "BanOnFacility": BanOnFacility,
            "BanOnResource": BanOnResource,
            "BanOnVo": BanOnVo,
            "Candidate": Candidate,
            "Category": Category,
            "Consent": Consent,
            "ConsentHub": ConsentHub,
            "Destination": Destination,
            "ExtSource": ExtSource,
            "Facility": Facility,
            "Group": Group,
            "Host": Host,
            "Member": Member,
            "Owner": Owner,
            "Publication": Publication,
            "PublicationForGUI": PublicationForGUI,
            "PublicationSystem": PublicationSystem,
            "Resource": Resource,
            "ResourceTag": ResourceTag,
            "RichDestination": RichDestination,
            "RichFacility": RichFacility,
            "RichGroup": RichGroup,
            "RichMember": RichMember,
            "RichResource": RichResource,
            "RichUser": RichUser,
            "SecurityTeam": SecurityTeam,
            "Service": Service,
            "ServiceForGUI": ServiceForGUI,
            "ServicesPackage": ServicesPackage,
            "TaskResult": TaskResult,
            "Thanks": Thanks,
            "ThanksForGUI": ThanksForGUI,
            "User": User,
            "UserExtSource": UserExtSource,
            "Vo": Vo,
        }
        if not val:
            return None
        return {"bean_name": val}

    attribute_map = {
        "id": "id",  # noqa: E501
        "bean_name": "beanName",  # noqa: E501
    }

    read_only_vars = {}

    _composed_schemas = {}

    @classmethod
    @convert_js_args_to_python_args
    def _from_openapi_data(cls, id, bean_name, *args, **kwargs):  # noqa: E501
        """PerunBean - a model defined in OpenAPI

        Args:
            id (int):
            bean_name (str):

        Keyword Args:
            _check_type (bool): if True, values for parameters in openapi_types
                                will be type checked and a TypeError will be
                                raised if the wrong type is input.
                                Defaults to True
            _path_to_item (tuple/list): This is a list of keys or values to
                                drill down to the model in received_data
                                when deserializing a response
            _spec_property_naming (bool): True if the variable names in the input data
                                are serialized names, as specified in the OpenAPI document.
                                False if the variable names in the input data
                                are pythonic names, e.g. snake case (default)
            _configuration (Configuration): the instance to use when
                                deserializing a file_type parameter.
                                If passed, type conversion is attempted
                                If omitted no type conversion is done.
            _visited_composed_classes (tuple): This stores a tuple of
                                classes that we have traveled through so that
                                if we see that class again we will not use its
                                discriminator again.
                                When traveling through a discriminator, the
                                composed schema that is
                                is traveled through is added to this set.
                                For example if Animal has a discriminator
                                petType and we pass in "Dog", and the class Dog
                                allOf includes Animal, we move through Animal
                                once using the discriminator, and pick Dog.
                                Then in Dog, we will make an instance of the
                                Animal class but this time we won't travel
                                through its discriminator because we passed in
                                _visited_composed_classes = (Animal,)
        """

        _check_type = kwargs.pop("_check_type", True)
        _spec_property_naming = kwargs.pop("_spec_property_naming", True)
        _path_to_item = kwargs.pop("_path_to_item", ())
        _configuration = kwargs.pop("_configuration", None)
        _visited_composed_classes = kwargs.pop("_visited_composed_classes", ())

        self = super(OpenApiModel, cls).__new__(cls)

        if args:
            for arg in args:
                if isinstance(arg, dict):
                    kwargs.update(arg)
                else:
                    raise ApiTypeError(
                        "Invalid positional arguments=%s passed to %s. Remove those invalid positional arguments."
                        % (
                            args,
                            self.__class__.__name__,
                        ),
                        path_to_item=_path_to_item,
                        valid_classes=(self.__class__,),
                    )

        self._data_store = {}
        self._check_type = _check_type
        self._spec_property_naming = _spec_property_naming
        self._path_to_item = _path_to_item
        self._configuration = _configuration
        self._visited_composed_classes = _visited_composed_classes + (self.__class__,)

        self.id = id
        self.bean_name = bean_name
        for var_name, var_value in kwargs.items():
            if (
                var_name not in self.attribute_map
                and self._configuration is not None
                and self._configuration.discard_unknown_keys
                and self.additional_properties_type is None
            ):
                # discard variable.
                continue
            setattr(self, var_name, var_value)
        return self

    required_properties = set(
        [
            "_data_store",
            "_check_type",
            "_spec_property_naming",
            "_path_to_item",
            "_configuration",
            "_visited_composed_classes",
        ]
    )

    @convert_js_args_to_python_args
    def __init__(self, id, bean_name, *args, **kwargs):  # noqa: E501
        """PerunBean - a model defined in OpenAPI

        Args:
            id (int):
            bean_name (str):

        Keyword Args:
            _check_type (bool): if True, values for parameters in openapi_types
                                will be type checked and a TypeError will be
                                raised if the wrong type is input.
                                Defaults to True
            _path_to_item (tuple/list): This is a list of keys or values to
                                drill down to the model in received_data
                                when deserializing a response
            _spec_property_naming (bool): True if the variable names in the input data
                                are serialized names, as specified in the OpenAPI document.
                                False if the variable names in the input data
                                are pythonic names, e.g. snake case (default)
            _configuration (Configuration): the instance to use when
                                deserializing a file_type parameter.
                                If passed, type conversion is attempted
                                If omitted no type conversion is done.
            _visited_composed_classes (tuple): This stores a tuple of
                                classes that we have traveled through so that
                                if we see that class again we will not use its
                                discriminator again.
                                When traveling through a discriminator, the
                                composed schema that is
                                is traveled through is added to this set.
                                For example if Animal has a discriminator
                                petType and we pass in "Dog", and the class Dog
                                allOf includes Animal, we move through Animal
                                once using the discriminator, and pick Dog.
                                Then in Dog, we will make an instance of the
                                Animal class but this time we won't travel
                                through its discriminator because we passed in
                                _visited_composed_classes = (Animal,)
        """

        _check_type = kwargs.pop("_check_type", True)
        _spec_property_naming = kwargs.pop("_spec_property_naming", False)
        _path_to_item = kwargs.pop("_path_to_item", ())
        _configuration = kwargs.pop("_configuration", None)
        _visited_composed_classes = kwargs.pop("_visited_composed_classes", ())

        if args:
            for arg in args:
                if isinstance(arg, dict):
                    kwargs.update(arg)
                else:
                    raise ApiTypeError(
                        "Invalid positional arguments=%s passed to %s. Remove those invalid positional arguments."
                        % (
                            args,
                            self.__class__.__name__,
                        ),
                        path_to_item=_path_to_item,
                        valid_classes=(self.__class__,),
                    )

        self._data_store = {}
        self._check_type = _check_type
        self._spec_property_naming = _spec_property_naming
        self._path_to_item = _path_to_item
        self._configuration = _configuration
        self._visited_composed_classes = _visited_composed_classes + (self.__class__,)

        self.id = id
        self.bean_name = bean_name
        for var_name, var_value in kwargs.items():
            if (
                var_name not in self.attribute_map
                and self._configuration is not None
                and self._configuration.discard_unknown_keys
                and self.additional_properties_type is None
            ):
                # discard variable.
                continue
            setattr(self, var_name, var_value)
            if var_name in self.read_only_vars:
                raise ApiAttributeError(
                    f"`{var_name}` is a read-only attribute. Use `from_openapi_data` to instantiate "
                    f"class with read only attributes."
                )



from unittest.mock import MagicMock, patch

import pytest

from perun.connector.connectors.LdapConnector import LdapConnector
from perun.connector.utils.ConfigStore import ConfigStore

loaded_config = ConfigStore.get_ldapc_config()
CONNECTOR = LdapConnector(loaded_config)

TEST_DATA = {
    "perunUserId": 1,
    "displayName": "Foe Toe",
    "cn": "Foe Toe",
    "memberOf": [
        "perunGroupId=1,perunVoId=1,dc=perun,dc=cesnet,dc=cz",
        "perunGroupId=2,perunVoId=1,dc=perun,dc=cesnet,dc=cz",
    ],
}
TEST_DATA_2 = {
    "perunUserId": 2,
    "displayName": "Joe Doe",
    "cn": "Joe Doe",
    "memberOf": [
        "perunGroupId=1,perunVoId=1,dc=perun,dc=cesnet,dc=cz",
        "perunGroupId=2,perunVoId=1,dc=perun,dc=cesnet,dc=cz",
    ],
}

TEST_ENTRIES = [TEST_DATA, TEST_DATA_2]
TEST_ENTRY = [TEST_DATA]
BASE = "ldap.base"
FILTERS = "(name=test_filters)"


@patch("perun.connector.connectors.LdapConnector.LdapConnector._search")
def test_search_for_entity_not_found(mock_request):
    CONNECTOR._search = MagicMock(return_value=None)
    result = CONNECTOR.search_for_entity(BASE, FILTERS)
    assert not result


@patch("perun.connector.connectors.LdapConnector.LdapConnector._search")
def test_search_for_entity_found(mock_request):
    CONNECTOR._search = MagicMock(return_value=TEST_ENTRY)
    result = CONNECTOR.search_for_entity(BASE, FILTERS)
    assert result == TEST_DATA


@patch("perun.connector.connectors.LdapConnector.LdapConnector._search")
def test_search_for_entity_found_invalid(mock_request):
    CONNECTOR._search = MagicMock(return_value=TEST_ENTRIES)
    expected_error_message = (
        "ldap_connector.search_for_entity - "
        "More than one entity found."
        + " query base:"
        + BASE
        + ", filter: "
        + FILTERS
        + "."
        + " Hint: Use method "
        "search_for_entities if "
        "you expect array of entities."
    )

    with pytest.raises(Exception) as error:
        _ = CONNECTOR.search_for_entity(BASE, FILTERS)

        assert str(error.value.args[0]) == expected_error_message


@patch("perun.connector.connectors.LdapConnector.LdapConnector._search")
def test_search_for_entities_not_found(mock_request):
    CONNECTOR._search = MagicMock(return_value=None)
    result = CONNECTOR.search_for_entities(BASE, FILTERS)
    assert not result


@patch("perun.connector.connectors.LdapConnector.LdapConnector._search")
def test_search_for_entities_found(mock_request):
    CONNECTOR._search = MagicMock(return_value=TEST_ENTRIES)
    result = CONNECTOR.search_for_entities(BASE, FILTERS)
    assert result == TEST_ENTRIES

    CONNECTOR._search = MagicMock(return_value=TEST_ENTRY)
    result = CONNECTOR.search_for_entities(BASE, FILTERS)
    assert result == TEST_ENTRY



import copy
import logging
from unittest.mock import MagicMock, patch

import pytest

import perun
from perun.connector.adapters.AdaptersManager import AdaptersManager
from perun.connector.adapters.LdapAdapter import AdapterSkipException, LdapAdapter
from perun.connector.adapters.PerunRpcAdapter import PerunRpcAdapter
from perun.connector.models.User import User
from perun.connector.perun_openapi.exceptions import ApiException
from perun.connector.utils.ConfigStore import ConfigStore


class HttpResponse:
    def __init__(self, input_data: str):
        self.status = None
        self.data = input_data
        self.reason = None

    def getheaders(self):
        return "headers"


BASE_MANAGER_CONFIG = {"adapters": []}

# 1st priority adapter is LDAP, 2nd priority is RPC - changes in priority
# might affect test results

LDAP_PRIORITY = 1
LDAP_CONFIG_DATA = {
    "type": "ldap",
    "priority": LDAP_PRIORITY,
    "username": "cn=admin,dc=muni,dc=cz",
    "base_dn": "dc=muni,dc=cz",
    "password": "mypassword",
    "start_tls": True,
    "servers": [
        {"hostname": "ldap://openldap", "port": 389},
        {"hostname": "ldap://openldap2", "port": 389},
    ],
}
LDAP_ADAPTER = LdapAdapter(LDAP_CONFIG_DATA, ConfigStore.get_attribute_map())

RPC_PRIORITY = 2
RPC_CONFIG_DATA = {
    "type": "openApi",
    "priority": RPC_PRIORITY,
    "host": "https://perun.cesnet.cz/krb/rpc",
    "auth_type": "BasicAuth",
    "username": "username",
    "password": "mypasswd",
    "api_key": "your_api_key",
    "access_token": "your_bearer_token",
    "cookie_filepath": "cookie.txt",
}

UNSUPPORTED_CONFIG_DATA = {"type": "curl", "priority": 3}
SUPPORTED_CONFIG_DATA = [LDAP_CONFIG_DATA, RPC_CONFIG_DATA]


def is_present_rpc_adapter(manager: AdaptersManager) -> bool:
    has_correct_name = manager.adapters.get(RPC_PRIORITY).get("name") == "rpc_adapter"
    has_correct_instance = isinstance(
        manager.adapters.get(RPC_PRIORITY).get("adapter"), PerunRpcAdapter
    )
    return has_correct_instance and has_correct_name


def is_present_ldap_adapter(manager: AdaptersManager) -> bool:
    has_correct_name = manager.adapters.get(LDAP_PRIORITY).get("name") == "ldap_adapter"
    has_correct_instance = isinstance(
        manager.adapters.get(LDAP_PRIORITY).get("adapter"), LdapAdapter
    )
    return has_correct_instance and has_correct_name


SUPPORTED_ADAPTERS_VALIDATORS = [is_present_rpc_adapter, is_present_ldap_adapter]


def test_create_manager_with_ldap_adapter():
    config = copy.deepcopy(BASE_MANAGER_CONFIG)
    config["adapters"].append(LDAP_CONFIG_DATA)

    manager = AdaptersManager(config, ConfigStore.get_attribute_map())

    assert len(manager.adapters) == 1
    assert is_present_ldap_adapter(manager)


def test_create_manager_with_rpc_adapter():
    config = copy.deepcopy(BASE_MANAGER_CONFIG)
    config["adapters"].append(RPC_CONFIG_DATA)

    manager = AdaptersManager(config, ConfigStore.get_attribute_map())

    assert len(manager.adapters) == 1
    assert is_present_rpc_adapter(manager)


def test_create_manager_with_all_supported_adapters():
    config = copy.deepcopy(BASE_MANAGER_CONFIG)
    config["adapters"] = SUPPORTED_CONFIG_DATA

    manager = AdaptersManager(config, ConfigStore.get_attribute_map())

    assert len(manager.adapters) == len(SUPPORTED_CONFIG_DATA)
    for validator in SUPPORTED_ADAPTERS_VALIDATORS:
        assert validator(manager)


def test_create_manager_with_unsupported_adapter(caplog):
    config = copy.deepcopy(BASE_MANAGER_CONFIG)
    config["adapters"] = copy.deepcopy(SUPPORTED_CONFIG_DATA)
    config["adapters"].append(UNSUPPORTED_CONFIG_DATA)

    unsupported_adapter_type = UNSUPPORTED_CONFIG_DATA.get("type")
    unsupported_adapter_warning = (
        f"Config file includes unsupported adapter "
        f'type "{unsupported_adapter_type}"'
    )

    with caplog.at_level(logging.WARNING):
        manager = AdaptersManager(config, ConfigStore.get_attribute_map())

        assert unsupported_adapter_warning in caplog.text
        assert len(manager.adapters) == len(SUPPORTED_CONFIG_DATA)
        for validator in SUPPORTED_ADAPTERS_VALIDATORS:
            assert validator(manager)


@patch("perun.connector.adapters.LdapAdapter.LdapAdapter.get_perun_user")
def test_find_method_on_first_adapter_successfully_execute(mock_request_1):
    config = copy.deepcopy(BASE_MANAGER_CONFIG)
    config["adapters"] = SUPPORTED_CONFIG_DATA

    manager = AdaptersManager(config, ConfigStore.get_attribute_map())

    test_user = User(1, "John Doe")
    perun.connector.adapters.LdapAdapter.LdapAdapter.get_perun_user = MagicMock(
        return_value=test_user
    )

    result = manager.get_perun_user("1", ["John Doe"])

    assert result == test_user


@patch("perun.connector.adapters.LdapAdapter.LdapAdapter.get_perun_user")
@patch("perun.connector.adapters.PerunRpcAdapter.PerunRpcAdapter.get_perun_user")
def test_find_method_on_second_adapter_successfully_execute(
    mock_request_1, mock_request_2, caplog
):
    config = copy.deepcopy(BASE_MANAGER_CONFIG)
    config["adapters"] = SUPPORTED_CONFIG_DATA

    manager = AdaptersManager(config, ConfigStore.get_attribute_map())

    unsupported_method_call_warning = (
        'Adapter not able to execute given action. Method: "get_perun_user" '
        "Adapter: ldap_adapter Going to try another adapter if available."
    )
    perun.connector.adapters.LdapAdapter.LdapAdapter.get_perun_user = MagicMock(
        side_effect=AdapterSkipException
    )

    test_user = User(1, "John Doe")
    perun.connector.adapters.PerunRpcAdapter.PerunRpcAdapter.get_perun_user = MagicMock(
        return_value=test_user
    )

    with caplog.at_level(logging.WARNING):
        result = manager.get_perun_user("1", ["John Doe"])
        assert unsupported_method_call_warning in caplog.text
        assert result == test_user


@patch("perun.connector.adapters.LdapAdapter.LdapAdapter.get_perun_user")
def test_found_method_fail_on_not_found_exception(mock_request_1, caplog):
    config = copy.deepcopy(BASE_MANAGER_CONFIG)
    config["adapters"] = SUPPORTED_CONFIG_DATA

    manager = AdaptersManager(config, ConfigStore.get_attribute_map())

    perun.connector.adapters.LdapAdapter.LdapAdapter.get_perun_user = MagicMock(
        side_effect=ApiException(
            http_resp=HttpResponse('"name":"UserNotExistsException"')
        )
    )

    # test logging of the exception
    expected_error_message = '"name":"UserNotExistsException"'

    with pytest.raises(Exception) as error:
        manager.get_perun_user("1", ["John Doe"])
        assert str(error.value.args[0]) == expected_error_message


@patch("perun.connector.adapters.LdapAdapter.LdapAdapter.get_perun_user")
@patch("perun.connector.adapters.PerunRpcAdapter.PerunRpcAdapter.get_perun_user")
def test_method_not_found_on_any_adapter(mock_request_1, mock_request_2):
    config = copy.deepcopy(BASE_MANAGER_CONFIG)
    config["adapters"] = SUPPORTED_CONFIG_DATA

    manager = AdaptersManager(config, ConfigStore.get_attribute_map())

    unsupported_method_call_message = (
        "'LdapAdapter' object has no attribute" " 'unsupported_method'"
    )
    unsupported_method_error = AttributeError(unsupported_method_call_message)

    perun.connector.adapters.LdapAdapter.LdapAdapter.get_perun_user = MagicMock(
        side_effect=unsupported_method_error
    )

    perun.connector.adapters.PerunRpcAdapter.PerunRpcAdapter.get_perun_user = MagicMock(
        side_effect=unsupported_method_error
    )

    method_not_found_on_any_adapter_message = (
        "None of the provided adapters "
        "was able to resolve method "
        '"unsupported_method"'
    )
    with pytest.raises(Exception) as error:
        _ = manager.get_perun_user("1", ["John Doe"])
        assert str(error.value.args[0]) == method_not_found_on_any_adapter_message



from perun.connector.models.HasIdAbstract import HasIdAbstract
from perun.connector.models.VO import VO


class Group(HasIdAbstract):
    def __init__(
        self, id: int, vo: VO, uuid: str, name: str, unique_name: str, description: str
    ):
        super().__init__(id)
        self.vo = vo
        self.uuid = uuid
        self.name = name
        self.unique_name = unique_name
        self.description = description

    def __str__(self):
        return (
            f"id: {self.id} vo: {self.vo} uuid: {self.uuid} name: "
            f"{self.name} unique_name: {self.unique_name}, descri"
            f"ption: {self.description}"
        )

    def __eq__(self, other):
        if isinstance(other, Group):
            return (
                self.id == other.id
                and self.vo == other.vo
                and self.uuid == other.uuid
                and self.name == other.name
                and self.unique_name == other.unique_name
                and self.description == other.description
            )
        return False

    def __hash__(self):
        return hash(
            (self.id, self.vo, self.uuid, self.name, self.unique_name, self.description)
        )
